from bbot.modules.base import BaseModule


class IP2Location(BaseModule):
    """
    IP2Location.io Geolocation API.
    """

    watched_events = ["IP_ADDRESS"]
    produced_events = ["GEOLOCATION"]
    flags = ["passive", "safe"]
    meta = {"description": "Query IP2location.io's API for geolocation information. ", "auth_required": True}
    options = {"api_key": "", "lang": ""}
    options_desc = {
        "api_key": "IP2location.io API Key",
        "lang": "Translation information(ISO639-1). The translation is only applicable for continent, country, region and city name.",
    }
    scope_distance_modifier = 1
    _priority = 2
    suppress_dupes = False

    base_url = "http://api.ip2location.io"

    async def setup(self):
        self.lang = self.config.get("lang", "")
        return await self.require_api_key()

    async def ping(self):
        url = self.build_url("8.8.8.8")
        r = await self.request_with_fail_count(url)
        resp_content = getattr(r, "text", "")
        assert getattr(r, "status_code", 0) == 200, resp_content

    def build_url(self, data):
        url = f"{self.base_url}/?key={self.api_key}&ip={data}&format=json&source=bbot"
        if self.lang:
            url = f"{url}&lang={self.lang}"
        return url

    async def handle_event(self, event):
        try:
            url = self.build_url(event.data)
            result = await self.request_with_fail_count(url)
            if result:
                geo_data = result.json()
                if not geo_data:
                    self.verbose(f"No JSON response from {url}")
            else:
                self.verbose(f"No response from {url}")
        except Exception:
            self.verbose(f"Error retrieving results for {event.data}", trace=True)
            return

        geo_data = {k: v for k, v in geo_data.items() if v is not None}
        if "error" in geo_data:
            error_msg = geo_data.get("error").get("error_message", "")
            if error_msg:
                self.warning(error_msg)
        elif geo_data:
            await self.emit_event(geo_data, "GEOLOCATION", event)



from .base import ModuleTestBase


class TestGithub_Org(ModuleTestBase):
    config_overrides = {"modules": {"github_org": {"api_key": "asdf"}}}
    modules_overrides = ["github_org", "speculate"]

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(url="https://api.github.com/zen")
        module_test.httpx_mock.add_response(
            url="https://api.github.com/orgs/blacklanternsecurity",
            json={
                "login": "blacklanternsecurity",
                "id": 25311592,
                "node_id": "MDEyOk9yZ2FuaXphdGlvbjI1MzExNTky",
                "url": "https://api.github.com/orgs/blacklanternsecurity",
                "repos_url": "https://api.github.com/orgs/blacklanternsecurity/repos",
                "events_url": "https://api.github.com/orgs/blacklanternsecurity/events",
                "hooks_url": "https://api.github.com/orgs/blacklanternsecurity/hooks",
                "issues_url": "https://api.github.com/orgs/blacklanternsecurity/issues",
                "members_url": "https://api.github.com/orgs/blacklanternsecurity/members{/member}",
                "public_members_url": "https://api.github.com/orgs/blacklanternsecurity/public_members{/member}",
                "avatar_url": "https://avatars.githubusercontent.com/u/25311592?v=4",
                "description": "Security Organization",
                "name": "Black Lantern Security",
                "company": None,
                "blog": "www.blacklanternsecurity.com",
                "location": "Charleston, SC",
                "email": None,
                "twitter_username": None,
                "is_verified": False,
                "has_organization_projects": True,
                "has_repository_projects": True,
                "public_repos": 70,
                "public_gists": 0,
                "followers": 415,
                "following": 0,
                "html_url": "https://github.com/blacklanternsecurity",
                "created_at": "2017-01-24T00:14:46Z",
                "updated_at": "2022-03-28T11:39:03Z",
                "archived_at": None,
                "type": "Organization",
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/orgs/blacklanternsecurity/repos?per_page=100&page=1",
            json=[
                {
                    "id": 459780477,
                    "node_id": "R_kgDOG2exfQ",
                    "name": "test_keys",
                    "full_name": "blacklanternsecurity/test_keys",
                    "private": False,
                    "owner": {
                        "login": "blacklanternsecurity",
                        "id": 79229934,
                        "node_id": "MDEyOk9yZ2FuaXphdGlvbjc5MjI5OTM0",
                        "avatar_url": "https://avatars.githubusercontent.com/u/79229934?v=4",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/blacklanternsecurity",
                        "html_url": "https://github.com/blacklanternsecurity",
                        "followers_url": "https://api.github.com/users/blacklanternsecurity/followers",
                        "following_url": "https://api.github.com/users/blacklanternsecurity/following{/other_user}",
                        "gists_url": "https://api.github.com/users/blacklanternsecurity/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/blacklanternsecurity/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/blacklanternsecurity/subscriptions",
                        "organizations_url": "https://api.github.com/users/blacklanternsecurity/orgs",
                        "repos_url": "https://api.github.com/users/blacklanternsecurity/repos",
                        "events_url": "https://api.github.com/users/blacklanternsecurity/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/blacklanternsecurity/received_events",
                        "type": "Organization",
                        "site_admin": False,
                    },
                    "html_url": "https://github.com/blacklanternsecurity/test_keys",
                    "description": None,
                    "fork": False,
                    "url": "https://api.github.com/repos/blacklanternsecurity/test_keys",
                    "forks_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/forks",
                    "keys_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/keys{/key_id}",
                    "collaborators_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/collaborators{/collaborator}",
                    "teams_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/teams",
                    "hooks_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/hooks",
                    "issue_events_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues/events{/number}",
                    "events_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/events",
                    "assignees_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/assignees{/user}",
                    "branches_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/branches{/branch}",
                    "tags_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/tags",
                    "blobs_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/blobs{/sha}",
                    "git_tags_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/tags{/sha}",
                    "git_refs_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/refs{/sha}",
                    "trees_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/trees{/sha}",
                    "statuses_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/statuses/{sha}",
                    "languages_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/languages",
                    "stargazers_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/stargazers",
                    "contributors_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/contributors",
                    "subscribers_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/subscribers",
                    "subscription_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/subscription",
                    "commits_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/commits{/sha}",
                    "git_commits_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/commits{/sha}",
                    "comments_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/comments{/number}",
                    "issue_comment_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues/comments{/number}",
                    "contents_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/contents/{+path}",
                    "compare_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/compare/{base}...{head}",
                    "merges_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/merges",
                    "archive_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/{archive_format}{/ref}",
                    "downloads_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/downloads",
                    "issues_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues{/number}",
                    "pulls_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/pulls{/number}",
                    "milestones_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/milestones{/number}",
                    "notifications_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/notifications{?since,all,participating}",
                    "labels_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/labels{/name}",
                    "releases_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/releases{/id}",
                    "deployments_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/deployments",
                    "created_at": "2022-02-15T23:10:51Z",
                    "updated_at": "2023-09-02T12:20:13Z",
                    "pushed_at": "2023-10-19T02:56:46Z",
                    "git_url": "git://github.com/blacklanternsecurity/test_keys.git",
                    "ssh_url": "git@github.com:blacklanternsecurity/test_keys.git",
                    "clone_url": "https://github.com/blacklanternsecurity/test_keys.git",
                    "svn_url": "https://github.com/blacklanternsecurity/test_keys",
                    "homepage": None,
                    "size": 2,
                    "stargazers_count": 2,
                    "watchers_count": 2,
                    "language": None,
                    "has_issues": True,
                    "has_projects": True,
                    "has_downloads": True,
                    "has_wiki": True,
                    "has_pages": False,
                    "has_discussions": False,
                    "forks_count": 32,
                    "mirror_url": None,
                    "archived": False,
                    "disabled": False,
                    "open_issues_count": 2,
                    "license": None,
                    "allow_forking": True,
                    "is_template": False,
                    "web_commit_signoff_required": False,
                    "topics": [],
                    "visibility": "public",
                    "forks": 32,
                    "open_issues": 2,
                    "watchers": 2,
                    "default_branch": "main",
                    "permissions": {"admin": False, "maintain": False, "push": False, "triage": False, "pull": True},
                }
            ],
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/orgs/blacklanternsecurity/members?per_page=100&page=1",
            json=[
                {
                    "login": "TheTechromancer",
                    "id": 20261699,
                    "node_id": "MDQ6VXNlcjIwMjYxNjk5",
                    "avatar_url": "https://avatars.githubusercontent.com/u/20261699?v=4",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/TheTechromancer",
                    "html_url": "https://github.com/TheTechromancer",
                    "followers_url": "https://api.github.com/users/TheTechromancer/followers",
                    "following_url": "https://api.github.com/users/TheTechromancer/following{/other_user}",
                    "gists_url": "https://api.github.com/users/TheTechromancer/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/TheTechromancer/starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/TheTechromancer/subscriptions",
                    "organizations_url": "https://api.github.com/users/TheTechromancer/orgs",
                    "repos_url": "https://api.github.com/users/TheTechromancer/repos",
                    "events_url": "https://api.github.com/users/TheTechromancer/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/TheTechromancer/received_events",
                    "type": "User",
                    "site_admin": False,
                }
            ],
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/users/TheTechromancer/repos?per_page=100&page=1",
            json=[
                {
                    "id": 688270318,
                    "node_id": "R_kgDOKQYr7g",
                    "name": "websitedemo",
                    "full_name": "TheTechromancer/websitedemo",
                    "private": False,
                    "owner": {
                        "login": "TheTechromancer",
                        "id": 20261699,
                        "node_id": "MDQ6VXNlcjIwMjYxNjk5",
                        "avatar_url": "https://avatars.githubusercontent.com/u/20261699?v=4",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/TheTechromancer",
                        "html_url": "https://github.com/TheTechromancer",
                        "followers_url": "https://api.github.com/users/TheTechromancer/followers",
                        "following_url": "https://api.github.com/users/TheTechromancer/following{/other_user}",
                        "gists_url": "https://api.github.com/users/TheTechromancer/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/TheTechromancer/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/TheTechromancer/subscriptions",
                        "organizations_url": "https://api.github.com/users/TheTechromancer/orgs",
                        "repos_url": "https://api.github.com/users/TheTechromancer/repos",
                        "events_url": "https://api.github.com/users/TheTechromancer/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/TheTechromancer/received_events",
                        "type": "User",
                        "site_admin": False,
                    },
                    "html_url": "https://github.com/TheTechromancer/websitedemo",
                    "description": None,
                    "fork": False,
                    "url": "https://api.github.com/repos/TheTechromancer/websitedemo",
                    "forks_url": "https://api.github.com/repos/TheTechromancer/websitedemo/forks",
                    "keys_url": "https://api.github.com/repos/TheTechromancer/websitedemo/keys{/key_id}",
                    "collaborators_url": "https://api.github.com/repos/TheTechromancer/websitedemo/collaborators{/collaborator}",
                    "teams_url": "https://api.github.com/repos/TheTechromancer/websitedemo/teams",
                    "hooks_url": "https://api.github.com/repos/TheTechromancer/websitedemo/hooks",
                    "issue_events_url": "https://api.github.com/repos/TheTechromancer/websitedemo/issues/events{/number}",
                    "events_url": "https://api.github.com/repos/TheTechromancer/websitedemo/events",
                    "assignees_url": "https://api.github.com/repos/TheTechromancer/websitedemo/assignees{/user}",
                    "branches_url": "https://api.github.com/repos/TheTechromancer/websitedemo/branches{/branch}",
                    "tags_url": "https://api.github.com/repos/TheTechromancer/websitedemo/tags",
                    "blobs_url": "https://api.github.com/repos/TheTechromancer/websitedemo/git/blobs{/sha}",
                    "git_tags_url": "https://api.github.com/repos/TheTechromancer/websitedemo/git/tags{/sha}",
                    "git_refs_url": "https://api.github.com/repos/TheTechromancer/websitedemo/git/refs{/sha}",
                    "trees_url": "https://api.github.com/repos/TheTechromancer/websitedemo/git/trees{/sha}",
                    "statuses_url": "https://api.github.com/repos/TheTechromancer/websitedemo/statuses/{sha}",
                    "languages_url": "https://api.github.com/repos/TheTechromancer/websitedemo/languages",
                    "stargazers_url": "https://api.github.com/repos/TheTechromancer/websitedemo/stargazers",
                    "contributors_url": "https://api.github.com/repos/TheTechromancer/websitedemo/contributors",
                    "subscribers_url": "https://api.github.com/repos/TheTechromancer/websitedemo/subscribers",
                    "subscription_url": "https://api.github.com/repos/TheTechromancer/websitedemo/subscription",
                    "commits_url": "https://api.github.com/repos/TheTechromancer/websitedemo/commits{/sha}",
                    "git_commits_url": "https://api.github.com/repos/TheTechromancer/websitedemo/git/commits{/sha}",
                    "comments_url": "https://api.github.com/repos/TheTechromancer/websitedemo/comments{/number}",
                    "issue_comment_url": "https://api.github.com/repos/TheTechromancer/websitedemo/issues/comments{/number}",
                    "contents_url": "https://api.github.com/repos/TheTechromancer/websitedemo/contents/{+path}",
                    "compare_url": "https://api.github.com/repos/TheTechromancer/websitedemo/compare/{base}...{head}",
                    "merges_url": "https://api.github.com/repos/TheTechromancer/websitedemo/merges",
                    "archive_url": "https://api.github.com/repos/TheTechromancer/websitedemo/{archive_format}{/ref}",
                    "downloads_url": "https://api.github.com/repos/TheTechromancer/websitedemo/downloads",
                    "issues_url": "https://api.github.com/repos/TheTechromancer/websitedemo/issues{/number}",
                    "pulls_url": "https://api.github.com/repos/TheTechromancer/websitedemo/pulls{/number}",
                    "milestones_url": "https://api.github.com/repos/TheTechromancer/websitedemo/milestones{/number}",
                    "notifications_url": "https://api.github.com/repos/TheTechromancer/websitedemo/notifications{?since,all,participating}",
                    "labels_url": "https://api.github.com/repos/TheTechromancer/websitedemo/labels{/name}",
                    "releases_url": "https://api.github.com/repos/TheTechromancer/websitedemo/releases{/id}",
                    "deployments_url": "https://api.github.com/repos/TheTechromancer/websitedemo/deployments",
                    "created_at": "2023-09-07T02:18:28Z",
                    "updated_at": "2023-09-07T02:20:18Z",
                    "pushed_at": "2023-09-07T02:34:45Z",
                    "git_url": "git://github.com/TheTechromancer/websitedemo.git",
                    "ssh_url": "git@github.com:TheTechromancer/websitedemo.git",
                    "clone_url": "https://github.com/TheTechromancer/websitedemo.git",
                    "svn_url": "https://github.com/TheTechromancer/websitedemo",
                    "homepage": None,
                    "size": 1,
                    "stargazers_count": 0,
                    "watchers_count": 0,
                    "language": "HTML",
                    "has_issues": True,
                    "has_projects": True,
                    "has_downloads": True,
                    "has_wiki": True,
                    "has_pages": True,
                    "has_discussions": False,
                    "forks_count": 0,
                    "mirror_url": None,
                    "archived": False,
                    "disabled": False,
                    "open_issues_count": 0,
                    "license": None,
                    "allow_forking": True,
                    "is_template": False,
                    "web_commit_signoff_required": False,
                    "topics": [],
                    "visibility": "public",
                    "forks": 0,
                    "open_issues": 0,
                    "watchers": 0,
                    "default_branch": "main",
                }
            ],
        )

    def check(self, module_test, events):
        assert len(events) == 6
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "blacklanternsecurity.com" and e.scope_distance == 0
            ]
        ), "Failed to emit target DNS_NAME"
        assert 1 == len(
            [e for e in events if e.type == "ORG_STUB" and e.data == "blacklanternsecurity" and e.scope_distance == 1]
        ), "Failed to find ORG_STUB"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "blacklanternsecurity"
                and "github-org" in e.tags
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity github"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "TheTechromancer"
                and "github-org-member" in e.tags
                and e.scope_distance == 2
            ]
        ), "Failed to find TheTechromancer github"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and e.data["url"] == "https://github.com/blacklanternsecurity/test_keys"
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity github repo"


class TestGithub_Org_No_Members(TestGithub_Org):
    config_overrides = {"modules": {"github_org": {"include_members": False}}}

    def check(self, module_test, events):
        assert len(events) == 5
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "blacklanternsecurity"
                and "github-org" in e.tags
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity github"
        assert 0 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "TheTechromancer"
            ]
        ), "Found TheTechromancer github"


class TestGithub_Org_MemberRepos(TestGithub_Org):
    config_overrides = {"modules": {"github_org": {"include_member_repos": True}}}

    def check(self, module_test, events):
        assert len(events) == 7
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and e.data["url"] == "https://github.com/TheTechromancer/websitedemo"
                and e.scope_distance == 2
            ]
        ), "Found to find TheTechromancer github repo"



import json
from werkzeug import Response

from .base import ModuleTestBase


class TestDastardly(ModuleTestBase):
    targets = ["http://127.0.0.1:5556/"]
    modules_overrides = ["httpx", "dastardly"]

    web_response = """<!DOCTYPE html>
    <html>
    <body>
        <a href="/test?test=yes">visit this<a/>
    </body>
    </html>"""

    def xss_handler(self, request):
        response = f"""<!DOCTYPE html>
    <html>
    <head>
        <title>Email Form</title>
    </head>
    <body>
        {request.args.get("test", "")}
    </body>
    </html>"""
        return Response(response, content_type="text/html")

    async def get_docker_ip(self, module_test):
        docker_ip = "172.17.0.1"
        try:
            ip_output = await module_test.scan.helpers.run(["ip", "-j", "-4", "a", "show", "dev", "docker0"])
            interface_json = json.loads(ip_output.stdout)
            docker_ip = interface_json[0]["addr_info"][0]["local"]
        except Exception:
            pass
        return docker_ip

    async def setup_after_prep(self, module_test):
        httpserver = module_test.request_fixture.getfixturevalue("bbot_httpserver_allinterfaces")
        httpserver.expect_request("/").respond_with_data(self.web_response)
        httpserver.expect_request("/test").respond_with_handler(self.xss_handler)

        # get docker IP
        docker_ip = await self.get_docker_ip(module_test)
        module_test.scan.target.add_target(docker_ip)

        # replace 127.0.0.1 with docker host IP to allow dastardly access to local http server
        old_filter_event = module_test.module.filter_event

        def new_filter_event(event):
            self.new_url = f"http://{docker_ip}:5556/"
            event.data["url"] = self.new_url
            event.parsed = module_test.scan.helpers.urlparse(self.new_url)
            return old_filter_event(event)

        module_test.monkeypatch.setattr(module_test.module, "filter_event", new_filter_event)

    def check(self, module_test, events):
        assert 1 == len([e for e in events if e.type == "VULNERABILITY"])
        assert 1 == len(
            [e for e in events if e.type == "VULNERABILITY" and f"{self.new_url}test" in e.data["description"]]
        )



from bbot.core.helpers.regexes import url_regexes

from .base import BaseModule


class OAUTH(BaseModule):
    watched_events = ["DNS_NAME", "URL_UNVERIFIED"]
    produced_events = ["DNS_NAME"]
    flags = ["affiliates", "subdomain-enum", "cloud-enum", "web-basic", "web-thorough", "active", "safe"]
    meta = {"description": "Enumerate OAUTH and OpenID Connect services"}
    options = {"try_all": False}
    options_desc = {"try_all": "Check for OAUTH/IODC on every subdomain and URL."}

    in_scope_only = False
    scope_distance_modifier = 1
    _max_event_handlers = 2

    async def setup(self):
        self.processed = set()
        self.regexes = list(url_regexes) + list(self.scan.dns_regexes)
        self.try_all = self.config.get("try_all", False)
        return True

    async def filter_event(self, event):
        if event.module == self or any(t in event.tags for t in ("target", "domain", "ms-auth-url")):
            return True
        elif self.try_all and event.scope_distance == 0:
            return True
        return False

    async def handle_event(self, event):
        _, domain = self.helpers.split_domain(event.data)
        source_domain = getattr(event, "source_domain", domain)
        if not self.scan.in_scope(source_domain):
            return

        oidc_tasks = []
        if event.scope_distance == 0:
            domain_hash = hash(domain)
            if domain_hash not in self.processed:
                self.processed.add(domain_hash)
                oidc_tasks.append(self.helpers.create_task(self.getoidc(f"https://login.windows.net/{domain}")))

        if event.type == "URL_UNVERIFIED":
            url = event.data
        else:
            url = f"https://{event.data}"

        oauth_tasks = []
        if self.try_all or any(t in event.tags for t in ("oauth-token-endpoint",)):
            oauth_tasks.append(self.helpers.create_task(self.getoauth(url)))
        if self.try_all or any(t in event.tags for t in ("ms-auth-url",)):
            for u in self.url_and_base(url):
                oidc_tasks.append(self.helpers.create_task(self.getoidc(u)))

        for oidc_task in oidc_tasks:
            url, token_endpoint, oidc_results = await oidc_task
            if token_endpoint:
                finding_event = self.make_event(
                    {
                        "description": f"OpenID Connect Endpoint (domain: {source_domain}) found at {url}",
                        "host": event.host,
                        "url": url,
                    },
                    "FINDING",
                    source=event,
                )
                finding_event.source_domain = source_domain
                await self.emit_event(finding_event)
                url_event = self.make_event(
                    token_endpoint, "URL_UNVERIFIED", source=event, tags=["affiliate", "oauth-token-endpoint"]
                )
                url_event.source_domain = source_domain
                await self.emit_event(url_event)
            for result in oidc_results:
                if result not in (domain, event.data):
                    event_type = "URL_UNVERIFIED" if self.helpers.is_url(result) else "DNS_NAME"
                    await self.emit_event(result, event_type, source=event, tags=["affiliate"])

        for oauth_task in oauth_tasks:
            url = await oauth_task
            if url:
                oauth_finding = self.make_event(
                    {
                        "description": f"Potentially Sprayable OAUTH Endpoint (domain: {source_domain}) at {url}",
                        "host": event.host,
                        "url": url,
                    },
                    "FINDING",
                    source=event,
                )
                oauth_finding.source_domain = source_domain
                await self.emit_event(oauth_finding)

    def url_and_base(self, url):
        yield url
        parsed = self.helpers.urlparse(url)
        baseurl = f"{parsed.scheme}://{parsed.netloc}/"
        if baseurl != url:
            yield baseurl

    async def getoidc(self, url):
        results = set()
        if not url.endswith("openid-configuration"):
            url = url.strip("/") + "/.well-known/openid-configuration"
        url_hash = hash("OIDC:" + url)
        token_endpoint = ""
        if url_hash not in self.processed:
            self.processed.add(url_hash)
            r = await self.helpers.request(url)
            if r is None:
                return url, token_endpoint, results
            try:
                json = r.json()
            except Exception:
                return url, token_endpoint, results
            if json and isinstance(json, dict):
                token_endpoint = json.get("token_endpoint", "")
                for found in self.helpers.search_dict_values(json, *self.regexes):
                    results.add(found)
        results -= {token_endpoint}
        return url, token_endpoint, results

    async def getoauth(self, url):
        data = {
            "grant_type": "authorization_code",
            "client_id": "xxx",
            "redirect_uri": "https://example.com",
            "code": "xxx",
            "client_secret": "xxx",
        }
        url_hash = hash("OAUTH:" + url)
        if url_hash not in self.processed:
            self.processed.add(url_hash)
            r = await self.helpers.request(url, method="POST", data=data)
            if r is None:
                return
            if r.status_code in (400, 401):
                if "json" in r.headers.get("content-type", "").lower():
                    if any(x in r.text.lower() for x in ("invalid_grant", "invalid_client")):
                        return url



import logging
import xmltodict
from deepdiff import DeepDiff
from contextlib import suppress
from xml.parsers.expat import ExpatError
from bbot.core.errors import HttpCompareError

log = logging.getLogger("bbot.core.helpers.diff")


class HttpCompare:
    def __init__(self, baseline_url, parent_helper, method="GET", allow_redirects=False, include_cache_buster=True):
        self.parent_helper = parent_helper
        self.baseline_url = baseline_url
        self.include_cache_buster = include_cache_buster
        self.method = method
        self.allow_redirects = allow_redirects
        self._baselined = False

    async def _baseline(self):
        if not self._baselined:
            # vanilla URL
            if self.include_cache_buster:
                url_1 = self.parent_helper.add_get_params(self.baseline_url, self.gen_cache_buster()).geturl()
            else:
                url_1 = self.baseline_url
            baseline_1 = await self.parent_helper.request(
                url_1, follow_redirects=self.allow_redirects, method=self.method
            )
            await self.parent_helper.sleep(1)
            # put random parameters in URL, headers, and cookies
            get_params = {self.parent_helper.rand_string(6): self.parent_helper.rand_string(6)}

            if self.include_cache_buster:
                get_params.update(self.gen_cache_buster())
            url_2 = self.parent_helper.add_get_params(self.baseline_url, get_params).geturl()
            baseline_2 = await self.parent_helper.request(
                url_2,
                headers={self.parent_helper.rand_string(6): self.parent_helper.rand_string(6)},
                cookies={self.parent_helper.rand_string(6): self.parent_helper.rand_string(6)},
                follow_redirects=self.allow_redirects,
                method=self.method,
            )

            self.baseline = baseline_1

            if baseline_1 is None or baseline_2 is None:
                log.debug("HTTP error while establishing baseline, aborting")
                raise HttpCompareError(
                    f"Can't get baseline from source URL: {url_1}:{baseline_1} / {url_2}:{baseline_2}"
                )
            if baseline_1.status_code != baseline_2.status_code:
                log.debug("Status code not stable during baseline, aborting")
                raise HttpCompareError("Can't get baseline from source URL")
            try:
                baseline_1_json = xmltodict.parse(baseline_1.text)
                baseline_2_json = xmltodict.parse(baseline_2.text)
            except ExpatError:
                log.debug(f"Cant HTML parse for {self.baseline_url}. Switching to text parsing as a backup")
                baseline_1_json = baseline_1.text.split("\n")
                baseline_2_json = baseline_2.text.split("\n")

            ddiff = DeepDiff(baseline_1_json, baseline_2_json, ignore_order=True, view="tree")
            self.ddiff_filters = []

            for k, v in ddiff.items():
                for x in list(ddiff[k]):
                    log.debug(f"Added {k} filter for path: {x.path()}")
                    self.ddiff_filters.append(x.path())

            self.baseline_json = baseline_1_json

            self.baseline_ignore_headers = [
                h.lower()
                for h in [
                    "date",
                    "last-modified",
                    "content-length",
                    "ETag",
                    "X-Pad",
                    "X-Backside-Transport",
                ]
            ]
            dynamic_headers = self.compare_headers(baseline_1.headers, baseline_2.headers)

            self.baseline_ignore_headers += [x.lower() for x in dynamic_headers]
            self.baseline_body_distance = self.compare_body(baseline_1_json, baseline_2_json)
            self._baselined = True

    def gen_cache_buster(self):
        return {self.parent_helper.rand_string(6): "1"}

    def compare_headers(self, headers_1, headers_2):
        differing_headers = []

        for i, headers in enumerate((headers_1, headers_2)):
            for header, value in list(headers.items()):
                if header.lower() in self.baseline_ignore_headers:
                    with suppress(KeyError):
                        log.debug(f'found ignored header "{header}" in headers_{i+1} and removed')
                        del headers[header]

        ddiff = DeepDiff(headers_1, headers_2, ignore_order=True, view="tree")

        for k, v in ddiff.items():
            for x in list(ddiff[k]):
                try:
                    header_value = str(x).split("'")[1]
                except KeyError:
                    continue
                differing_headers.append(header_value)
        return differing_headers

    def compare_body(self, content_1, content_2):
        if content_1 == content_2:
            return True

        ddiff = DeepDiff(content_1, content_2, ignore_order=True, view="tree", exclude_paths=self.ddiff_filters)

        if len(ddiff.keys()) == 0:
            return True
        else:
            log.debug(ddiff)
            return False

    async def compare(
        self, subject, headers=None, cookies=None, check_reflection=False, method="GET", allow_redirects=False
    ):
        """
        Compares a URL with the baseline, with optional headers or cookies added

        Returns (match (bool), reason (str), reflection (bool),subject_response (requests response object))
            where "match" is whether the content matched against the baseline, and
                "reason" is the location of the change ("code", "body", "header", or None), and
                "reflection" is whether the value was reflected in the HTTP response
        """
        await self._baseline()

        reflection = False
        if self.include_cache_buster:
            cache_key, cache_value = list(self.gen_cache_buster().items())[0]
            url = self.parent_helper.add_get_params(subject, {cache_key: cache_value}).geturl()
        else:
            url = subject
        subject_response = await self.parent_helper.request(
            url, headers=headers, cookies=cookies, follow_redirects=allow_redirects, method=method
        )

        if subject_response is None:
            # this can be caused by a WAF not liking the header, so we really arent interested in it
            return (True, "403", reflection, subject_response)

        if check_reflection:
            for arg in (headers, cookies):
                if arg is not None:
                    for k, v in arg.items():
                        if v in subject_response.text:
                            reflection = True
                            break

            subject_params = self.parent_helper.get_get_params(subject)
            for k, v in subject_params.items():
                if k != cache_key:
                    for item in v:
                        if item in subject_response.text:
                            reflection = True
                            break
        try:
            subject_json = xmltodict.parse(subject_response.text)

        except ExpatError:
            log.debug(f"Cant HTML parse for {subject.split('?')[0]}. Switching to text parsing as a backup")
            subject_json = subject_response.text.split("\n")

        diff_reasons = []

        if self.baseline.status_code != subject_response.status_code:
            log.debug(
                f"status code was different [{str(self.baseline.status_code)}] -> [{str(subject_response.status_code)}], no match"
            )
            diff_reasons.append("code")

        different_headers = self.compare_headers(self.baseline.headers, subject_response.headers)
        if different_headers:
            log.debug(f"headers were different, no match")
            diff_reasons.append("header")

        if self.compare_body(self.baseline_json, subject_json) == False:
            log.debug(f"difference in HTML body, no match")

            diff_reasons.append("body")

        if not diff_reasons:
            return (True, [], False, None)
        else:
            return (False, diff_reasons, reflection, subject_response)

    async def canary_check(self, url, mode, rounds=6):
        """
        test detection using a canary to find hosts giving bad results
        """
        await self._baseline()
        headers = None
        cookies = None
        for i in range(0, rounds):
            random_params = {self.parent_helper.rand_string(7): self.parent_helper.rand_string(6)}
            new_url = str(url)
            if mode == "getparam":
                new_url = self.parent_helper.add_get_params(url, random_params).geturl()
            elif mode == "header":
                headers = random_params
            elif mode == "cookie":
                cookies = random_params
            else:
                raise ValueError(f'Invalid mode: "{mode}", choose from: getparam, header, cookie')

            match, reasons, reflection, subject_response = await self.compare(
                new_url, headers=headers, cookies=cookies
            )

            # a nonsense header "caused" a difference, we need to abort
            if match == False:
                return False
        return True



import dns
import time
import asyncio
import logging
import ipaddress
import traceback
import contextlib
import dns.exception
import dns.asyncresolver
from cachetools import LRUCache
from contextlib import suppress

from .regexes import dns_name_regex
from bbot.core.helpers.ratelimiter import RateLimiter
from bbot.core.helpers.async_helpers import NamedLock
from bbot.core.errors import ValidationError, DNSError, DNSWildcardBreak
from .misc import is_ip, is_domain, is_dns_name, domain_parents, parent_domain, rand_string, cloudcheck

log = logging.getLogger("bbot.core.helpers.dns")


class BBOTAsyncResolver(dns.asyncresolver.Resolver):
    """Custom asynchronous resolver for BBOT with rate limiting.

    This class extends dnspython's async resolver and provides additional support for rate-limiting DNS queries.
    The maximum number of queries allowed per second can be customized via BBOT's config.

    Attributes:
        _parent_helper: A reference to the instantiated `ConfigAwareHelper` (typically `scan.helpers`).
        _dns_rate_limiter (RateLimiter): An instance of the RateLimiter class for DNS query rate-limiting.

    Args:
        *args: Positional arguments passed to the base resolver.
        **kwargs: Keyword arguments. '_parent_helper' is expected among these to provide configuration data for
                  rate-limiting. All other keyword arguments are passed to the base resolver.
    """

    def __init__(self, *args, **kwargs):
        self._parent_helper = kwargs.pop("_parent_helper")
        dns_queries_per_second = self._parent_helper.config.get("dns_queries_per_second", 100)
        self._dns_rate_limiter = RateLimiter(dns_queries_per_second, "DNS")
        super().__init__(*args, **kwargs)
        self.rotate = True

    async def resolve(self, *args, **kwargs):
        async with self._dns_rate_limiter:
            return await super().resolve(*args, **kwargs)


class DNSHelper:
    """Helper class for DNS-related operations within BBOT.

    This class provides mechanisms for host resolution, wildcard domain detection, event tagging, and more.
    It centralizes all DNS-related activities in BBOT, offering both synchronous and asynchronous methods
    for DNS resolution, as well as various utilities for batch resolution and DNS query filtering.

    Attributes:
        parent_helper: A reference to the instantiated `ConfigAwareHelper` (typically `scan.helpers`).
        resolver (BBOTAsyncResolver): An asynchronous DNS resolver tailored for BBOT with rate-limiting capabilities.
        timeout (int): The timeout value for DNS queries. Defaults to 5 seconds.
        retries (int): The number of retries for failed DNS queries. Defaults to 1.
        abort_threshold (int): The threshold for aborting after consecutive failed queries. Defaults to 50.
        max_dns_resolve_distance (int): Maximum allowed distance for DNS resolution. Defaults to 4.
        all_rdtypes (list): A list of DNS record types to be considered during operations.
        wildcard_ignore (tuple): Domains to be ignored during wildcard detection.
        wildcard_tests (int): Number of tests to be run for wildcard detection. Defaults to 5.
        _wildcard_cache (dict): Cache for wildcard detection results.
        _dns_cache (LRUCache): Cache for DNS resolution results, limited in size.
        _event_cache (LRUCache): Cache for event resolution results, tags. Limited in size.
        resolver_file (Path): File containing system's current resolver nameservers.
        filter_bad_ptrs (bool): Whether to filter out DNS names that appear to be auto-generated PTR records. Defaults to True.

    Args:
        parent_helper: The parent helper object with configuration details and utilities.

    Raises:
        DNSError: If an issue arises when creating the BBOTAsyncResolver instance.

    Examples:
        >>> dns_helper = DNSHelper(parent_config)
        >>> resolved_host = dns_helper.resolver.resolve("example.com")
    """

    all_rdtypes = ["A", "AAAA", "SRV", "MX", "NS", "SOA", "CNAME", "TXT"]

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        try:
            self.resolver = BBOTAsyncResolver(_parent_helper=self.parent_helper)
        except Exception as e:
            raise DNSError(f"Failed to create BBOT DNS resolver: {e}")
        self.timeout = self.parent_helper.config.get("dns_timeout", 5)
        self.retries = self.parent_helper.config.get("dns_retries", 1)
        self.abort_threshold = self.parent_helper.config.get("dns_abort_threshold", 50)
        self.max_dns_resolve_distance = self.parent_helper.config.get("max_dns_resolve_distance", 5)
        self.resolver.timeout = self.timeout
        self.resolver.lifetime = self.timeout
        self._resolver_list = None

        # skip certain queries
        dns_omit_queries = self.parent_helper.config.get("dns_omit_queries", None)
        if not dns_omit_queries:
            dns_omit_queries = []
        self.dns_omit_queries = dict()
        for d in dns_omit_queries:
            d = d.split(":")
            if len(d) == 2:
                rdtype, query = d
                rdtype = rdtype.upper()
                query = query.lower()
                try:
                    self.dns_omit_queries[rdtype].add(query)
                except KeyError:
                    self.dns_omit_queries[rdtype] = {query}

        self.wildcard_ignore = self.parent_helper.config.get("dns_wildcard_ignore", None)
        if not self.wildcard_ignore:
            self.wildcard_ignore = []
        self.wildcard_ignore = tuple([str(d).strip().lower() for d in self.wildcard_ignore])
        self.wildcard_tests = self.parent_helper.config.get("dns_wildcard_tests", 5)
        self._wildcard_cache = dict()
        # since wildcard detection takes some time, This is to prevent multiple
        # modules from kicking off wildcard detection for the same domain at the same time
        self._wildcard_lock = NamedLock()
        self._dns_connectivity_lock = asyncio.Lock()
        self._last_dns_success = None
        self._last_connectivity_warning = time.time()
        # keeps track of warnings issued for wildcard detection to prevent duplicate warnings
        self._dns_warnings = set()
        self._errors = dict()
        self.fallback_nameservers_file = self.parent_helper.wordlist_dir / "nameservers.txt"
        self._debug = self.parent_helper.config.get("dns_debug", False)
        self._dummy_modules = dict()
        self._dns_cache = LRUCache(maxsize=10000)
        self._event_cache = LRUCache(maxsize=10000)
        self._event_cache_locks = NamedLock()

        # for mocking DNS queries
        self._orig_resolve_raw = None
        self._mock_table = {}

        # copy the system's current resolvers to a text file for tool use
        self.system_resolvers = dns.resolver.Resolver().nameservers
        if len(self.system_resolvers) == 1:
            log.warning("BBOT performs better with multiple DNS servers. Your system currently only has one.")
        self.resolver_file = self.parent_helper.tempfile(self.system_resolvers, pipe=False)

        self.filter_bad_ptrs = self.parent_helper.config.get("dns_filter_ptrs", True)

    async def resolve(self, query, **kwargs):
        """Resolve DNS names and IP addresses to their corresponding results.

        This is a high-level function that can translate a given domain name to its associated IP addresses
        or an IP address to its corresponding domain names. It's structured for ease of use within modules
        and will abstract away most of the complexity of DNS resolution, returning a simple set of results.

        Args:
            query (str): The domain name or IP address to resolve.
            **kwargs: Additional arguments to be passed to the resolution process.

        Returns:
            set: A set containing resolved domain names or IP addresses.

        Examples:
            >>> results = await resolve("1.2.3.4")
            {"evilcorp.com"}

            >>> results = await resolve("evilcorp.com")
            {"1.2.3.4", "dead::beef"}
        """
        results = set()
        try:
            r = await self.resolve_raw(query, **kwargs)
            if r:
                raw_results, errors = r
                for rdtype, answers in raw_results:
                    for answer in answers:
                        for _, t in self.extract_targets(answer):
                            results.add(t)
        except BaseException:
            log.trace(f"Caught exception in resolve({query}, {kwargs}):")
            log.trace(traceback.format_exc())
            raise

        self.debug(f"Results for {query} with kwargs={kwargs}: {results}")
        return results

    async def resolve_raw(self, query, **kwargs):
        """Resolves the given query to its associated DNS records.

        This function is a foundational method for DNS resolution in this class. It understands both IP addresses and
        hostnames and returns their associated records in a raw format provided by the dnspython library.

        Args:
            query (str): The IP address or hostname to resolve.
            type (str or list[str], optional): Specifies the DNS record type(s) to fetch. Can be a single type like 'A'
                or a list like ['A', 'AAAA']. If set to 'any', 'all', or '*', it fetches all supported types. If not
                specified, the function defaults to fetching 'A' and 'AAAA' records.
            **kwargs: Additional arguments that might be passed to the resolver.

        Returns:
            tuple: A tuple containing two lists:
                - list: A list of tuples where each tuple consists of a record type string (like 'A') and the associated
                  raw dnspython answer.
                - list: A list of tuples where each tuple consists of a record type string and the associated error if
                  there was an issue fetching the record.

        Examples:
            >>> await resolve_raw("8.8.8.8")
            ([('PTR', <dns.resolver.Answer object at 0x7f4a47cdb1d0>)], [])

            >>> await resolve_raw("dns.google")
            ([('A', <dns.resolver.Answer object at 0x7f4a47ce46d0>), ('AAAA', <dns.resolver.Answer object at 0x7f4a47ce4710>)], [])
        """
        # DNS over TCP is more reliable
        # But setting this breaks DNS resolution on Ubuntu because systemd-resolve doesn't support TCP
        # kwargs["tcp"] = True
        results = []
        errors = []
        try:
            query = str(query).strip()
            if is_ip(query):
                kwargs.pop("type", None)
                kwargs.pop("rdtype", None)
                results, errors = await self._resolve_ip(query, **kwargs)
                return [("PTR", results)], [("PTR", e) for e in errors]
            else:
                types = ["A", "AAAA"]
                kwargs.pop("rdtype", None)
                if "type" in kwargs:
                    t = kwargs.pop("type")
                    types = self._parse_rdtype(t, default=types)
                for t in types:
                    r, e = await self._resolve_hostname(query, rdtype=t, **kwargs)
                    if r:
                        results.append((t, r))
                    for error in e:
                        errors.append((t, error))
        except BaseException:
            log.trace(f"Caught exception in resolve_raw({query}, {kwargs}):")
            log.trace(traceback.format_exc())
            raise

        return (results, errors)

    async def _resolve_hostname(self, query, **kwargs):
        """Translate a hostname into its corresponding IP addresses.

        This is the foundational function for converting a domain name into its associated IP addresses. It's designed
        for internal use within the class and handles retries, caching, and a variety of error/timeout scenarios.
        It also respects certain configurations that might ask to skip certain types of queries. Results are returned
        in the default dnspython answer object format.

        Args:
            query (str): The hostname to resolve.
            rdtype (str, optional): The type of DNS record to query (e.g., 'A', 'AAAA'). Defaults to 'A'.
            retries (int, optional): The number of times to retry on failure. Defaults to class-wide `retries`.
            use_cache (bool, optional): Whether to check the cache before trying a fresh resolution. Defaults to True.
            **kwargs: Additional arguments that might be passed to the resolver.

        Returns:
            tuple: A tuple containing:
                - list: A list of resolved IP addresses.
                - list: A list of errors encountered during the resolution process.

        Examples:
            >>> results, errors = await _resolve_hostname("google.com")
            (<dns.resolver.Answer object at 0x7f4a4b2caf50>, [])
        """
        self.debug(f"Resolving {query} with kwargs={kwargs}")
        results = []
        errors = []
        rdtype = kwargs.get("rdtype", "A")

        # skip certain queries if requested
        if rdtype in self.dns_omit_queries:
            if any(h == query or query.endswith(f".{h}") for h in self.dns_omit_queries[rdtype]):
                self.debug(f"Skipping {rdtype}:{query} because it's omitted in the config")
                return results, errors

        parent = self.parent_helper.parent_domain(query)
        retries = kwargs.pop("retries", self.retries)
        use_cache = kwargs.pop("use_cache", True)
        tries_left = int(retries) + 1
        parent_hash = hash(f"{parent}:{rdtype}")
        dns_cache_hash = hash(f"{query}:{rdtype}")
        while tries_left > 0:
            try:
                if use_cache:
                    results = self._dns_cache.get(dns_cache_hash, [])
                if not results:
                    error_count = self._errors.get(parent_hash, 0)
                    if error_count >= self.abort_threshold:
                        connectivity = await self._connectivity_check()
                        if connectivity:
                            log.verbose(
                                f'Aborting query "{query}" because failed {rdtype} queries for "{parent}" ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
                            )
                            if parent_hash not in self._dns_warnings:
                                log.verbose(
                                    f'Aborting future {rdtype} queries to "{parent}" because error count ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
                                )
                            self._dns_warnings.add(parent_hash)
                            return results, errors
                    results = await self._catch(self.resolver.resolve, query, **kwargs)
                    if use_cache:
                        self._dns_cache[dns_cache_hash] = results
                    if parent_hash in self._errors:
                        self._errors[parent_hash] = 0
                break
            except (
                dns.resolver.NoNameservers,
                dns.exception.Timeout,
                dns.resolver.LifetimeTimeout,
                TimeoutError,
            ) as e:
                try:
                    self._errors[parent_hash] += 1
                except KeyError:
                    self._errors[parent_hash] = 1
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    break
                tries_left -= 1
                err_msg = (
                    f'DNS error or timeout for {rdtype} query "{query}" ({self._errors[parent_hash]:,} so far): {e}'
                )
                if tries_left > 0:
                    retry_num = (retries + 1) - tries_left
                    self.debug(err_msg)
                    self.debug(f"Retry (#{retry_num}) resolving {query} with kwargs={kwargs}")
                else:
                    log.verbose(err_msg)

        if results:
            self._last_dns_success = time.time()
            self.debug(f"Answers for {query} with kwargs={kwargs}: {list(results)}")

        if errors:
            self.debug(f"Errors for {query} with kwargs={kwargs}: {errors}")

        return results, errors

    async def _resolve_ip(self, query, **kwargs):
        """Translate an IP address into a corresponding DNS name.

        This is the most basic function that will convert an IP address into its associated domain name. It handles
        retries, caching, and multiple types of timeout/error scenarios internally. The function is intended for
        internal use and should not be directly called by modules without understanding its intricacies.

        Args:
            query (str): The IP address to be reverse-resolved.
            retries (int, optional): The number of times to retry on failure. Defaults to 0.
            use_cache (bool, optional): Whether to check the cache for the result before attempting resolution. Defaults to True.
            **kwargs: Additional arguments to be passed to the resolution process.

        Returns:
            tuple: A tuple containing:
                - list: A list of resolved domain names (in default dnspython answer format).
                - list: A list of errors encountered during resolution.

        Examples:
            >>> results, errors = await _resolve_ip("8.8.8.8")
            (<dns.resolver.Answer object at 0x7f4a47cdb1d0>, [])
        """
        self.debug(f"Reverse-resolving {query} with kwargs={kwargs}")
        retries = kwargs.pop("retries", 0)
        use_cache = kwargs.pop("use_cache", True)
        tries_left = int(retries) + 1
        results = []
        errors = []
        dns_cache_hash = hash(f"{query}:PTR")
        while tries_left > 0:
            try:
                if use_cache:
                    results = self._dns_cache.get(dns_cache_hash, [])
                if not results:
                    results = await self._catch(self.resolver.resolve_address, query, **kwargs)
                    if use_cache:
                        self._dns_cache[dns_cache_hash] = results
                break
            except (
                dns.exception.Timeout,
                dns.resolver.LifetimeTimeout,
                dns.resolver.NoNameservers,
                TimeoutError,
            ) as e:
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    self.debug(f"{e} (query={query}, kwargs={kwargs})")
                    break
                else:
                    tries_left -= 1
                    if tries_left > 0:
                        retry_num = (retries + 2) - tries_left
                        self.debug(f"Retrying (#{retry_num}) {query} with kwargs={kwargs}")

        if results:
            self._last_dns_success = time.time()

        return results, errors

    async def handle_wildcard_event(self, event, children):
        """
        Used within BBOT's scan manager to detect and tag DNS wildcard events.

        Wildcards are detected for every major record type. If a wildcard is detected, its data
        is overwritten, for example: `_wildcard.evilcorp.com`.

        Args:
            event (object): The event to check for wildcards.
            children (list): A list of the event's resulting DNS children after resolution.

        Returns:
            None: This method modifies the `event` in place and does not return a value.

        Examples:
            >>> handle_wildcard_event(event, children)
            # The `event` might now have tags like ["wildcard", "a-wildcard", "aaaa-wildcard"] and
            # its `data` attribute might be modified to "_wildcard.evilcorp.com" if it was detected
            # as a wildcard.
        """
        log.debug(f"Entering handle_wildcard_event({event}, children={children})")
        try:
            event_host = str(event.host)
            # wildcard checks
            if not is_ip(event.host):
                # check if the dns name itself is a wildcard entry
                wildcard_rdtypes = await self.is_wildcard(event_host)
                for rdtype, (is_wildcard, wildcard_host) in wildcard_rdtypes.items():
                    wildcard_tag = "error"
                    if is_wildcard == True:
                        event.add_tag("wildcard")
                        wildcard_tag = "wildcard"
                    event.add_tag(f"{rdtype.lower()}-{wildcard_tag}")

            # wildcard event modification (www.evilcorp.com --> _wildcard.evilcorp.com)
            if not is_ip(event.host) and children:
                if wildcard_rdtypes:
                    # these are the rdtypes that successfully resolve
                    resolved_rdtypes = set([c.upper() for c in children])
                    # these are the rdtypes that have wildcards
                    wildcard_rdtypes_set = set(wildcard_rdtypes)
                    # consider the event a full wildcard if all its records are wildcards
                    event_is_wildcard = False
                    if resolved_rdtypes:
                        event_is_wildcard = all(r in wildcard_rdtypes_set for r in resolved_rdtypes)

                    if event_is_wildcard:
                        if event.type in ("DNS_NAME",) and not "_wildcard" in event.data.split("."):
                            wildcard_parent = self.parent_helper.parent_domain(event_host)
                            for rdtype, (_is_wildcard, _parent_domain) in wildcard_rdtypes.items():
                                if _is_wildcard:
                                    wildcard_parent = _parent_domain
                                    break
                            wildcard_data = f"_wildcard.{wildcard_parent}"
                            if wildcard_data != event.data:
                                log.debug(
                                    f'Wildcard detected, changing event.data "{event.data}" --> "{wildcard_data}"'
                                )
                                event.data = wildcard_data
                # tag wildcard domains for convenience
                elif is_domain(event_host) or hash(event_host) in self._wildcard_cache:
                    event_target = "target" in event.tags
                    wildcard_domain_results = await self.is_wildcard_domain(event_host, log_info=event_target)
                    for hostname, wildcard_domain_rdtypes in wildcard_domain_results.items():
                        if wildcard_domain_rdtypes:
                            event.add_tag("wildcard-domain")
                            for rdtype, ips in wildcard_domain_rdtypes.items():
                                event.add_tag(f"{rdtype.lower()}-wildcard-domain")
        finally:
            log.debug(f"Finished handle_wildcard_event({event}, children={children})")

    async def resolve_event(self, event, minimal=False):
        """
        Tag the given event with the appropriate DNS record types and optionally create child
        events based on DNS resolutions.

        Args:
            event (object): The event to be resolved and tagged.
            minimal (bool, optional): If set to True, the function will perform minimal DNS
                resolution. Defaults to False.

        Returns:
            tuple: A 4-tuple containing the following items:
                - event_tags (set): Set of tags for the event.
                - event_whitelisted (bool): Whether the event is whitelisted.
                - event_blacklisted (bool): Whether the event is blacklisted.
                - dns_children (dict): Dictionary containing child events from DNS resolutions.

        Examples:
            >>> event = make_event("evilcorp.com")
            >>> resolve_event(event)
            ({'resolved', 'ns-record', 'a-record',}, False, False, {'A': {IPv4Address('1.2.3.4'), IPv4Address('1.2.3.5')}, 'NS': {'ns1.evilcorp.com'}})

        Note:
            This method does not modify the passed in `event`. Instead, it returns data
            that can be used to modify or act upon the `event`.
        """
        log.debug(f"Resolving {event}")
        event_host = str(event.host)
        event_tags = set()
        dns_children = dict()
        event_whitelisted = False
        event_blacklisted = False

        try:
            if (not event.host) or (event.type in ("IP_RANGE",)):
                return event_tags, event_whitelisted, event_blacklisted, dns_children

            # lock to ensure resolution of the same host doesn't start while we're working here
            async with self._event_cache_locks.lock(event_host):
                # try to get data from cache
                _event_tags, _event_whitelisted, _event_blacklisted, _dns_children = self.event_cache_get(event_host)
                event_tags.update(_event_tags)
                # if we found it, return it
                if _event_whitelisted is not None:
                    return event_tags, _event_whitelisted, _event_blacklisted, _dns_children

                # then resolve
                types = ()
                if self.parent_helper.is_ip(event.host):
                    if not minimal:
                        types = ("PTR",)
                else:
                    if event.type == "DNS_NAME" and not minimal:
                        types = self.all_rdtypes
                    else:
                        types = ("A", "AAAA")

                if types:
                    for t in types:
                        resolved_raw, errors = await self.resolve_raw(event_host, type=t, use_cache=True)
                        for rdtype, e in errors:
                            if rdtype not in resolved_raw:
                                event_tags.add(f"{rdtype.lower()}-error")
                        for rdtype, records in resolved_raw:
                            rdtype = str(rdtype).upper()
                            if records:
                                event_tags.add("resolved")
                                event_tags.add(f"{rdtype.lower()}-record")

                            # whitelisting and blacklisting of IPs
                            for r in records:
                                for _, t in self.extract_targets(r):
                                    if t:
                                        ip = self.parent_helper.make_ip_type(t)

                                        if rdtype in ("A", "AAAA", "CNAME"):
                                            with contextlib.suppress(ValidationError):
                                                if self.parent_helper.is_ip(ip):
                                                    if self.parent_helper.scan.whitelisted(ip):
                                                        event_whitelisted = True
                                            with contextlib.suppress(ValidationError):
                                                if self.parent_helper.scan.blacklisted(ip):
                                                    event_blacklisted = True

                                        if self.filter_bad_ptrs and rdtype in ("PTR") and self.parent_helper.is_ptr(t):
                                            self.debug(f"Filtering out bad PTR: {t}")
                                            continue

                                        try:
                                            dns_children[rdtype].add(ip)
                                        except KeyError:
                                            dns_children[rdtype] = {ip}

                    # tag with cloud providers
                    if not self.parent_helper.in_tests:
                        to_check = set()
                        if event.type == "IP_ADDRESS":
                            to_check.add(event.data)
                        for rdtype, ips in dns_children.items():
                            if rdtype in ("A", "AAAA"):
                                for ip in ips:
                                    to_check.add(ip)
                        for ip in to_check:
                            provider, provider_type, subnet = cloudcheck(ip)
                            if provider:
                                event_tags.add(f"{provider_type}-{provider}")

                    # if needed, mark as unresolved
                    if not is_ip(event_host) and "resolved" not in event_tags:
                        event_tags.add("unresolved")
                    # check for private IPs
                    for rdtype, ips in dns_children.items():
                        for ip in ips:
                            try:
                                ip = ipaddress.ip_address(ip)
                                if ip.is_private:
                                    event_tags.add("private-ip")
                            except ValueError:
                                continue

                    self._event_cache[event_host] = (event_tags, event_whitelisted, event_blacklisted, dns_children)

            return event_tags, event_whitelisted, event_blacklisted, dns_children

        finally:
            log.debug(f"Finished resolving {event}")

    def event_cache_get(self, host):
        """
        Retrieves cached event data based on the given host.

        Args:
            host (str): The host for which the event data is to be retrieved.

        Returns:
            tuple: A 4-tuple containing the following items:
                - event_tags (set): Set of tags for the event.
                - event_whitelisted (bool or None): Whether the event is whitelisted. Returns None if not found.
                - event_blacklisted (bool or None): Whether the event is blacklisted. Returns None if not found.
                - dns_children (set): Set containing child events from DNS resolutions.

        Examples:
            Assuming an event with host "www.evilcorp.com" has been cached:

            >>> event_cache_get("www.evilcorp.com")
            ({"resolved", "a-record"}, False, False, {'1.2.3.4'})

            Assuming no event with host "www.notincache.com" has been cached:

            >>> event_cache_get("www.notincache.com")
            (set(), None, None, set())
        """
        try:
            event_tags, event_whitelisted, event_blacklisted, dns_children = self._event_cache[host]
            return (event_tags, event_whitelisted, event_blacklisted, dns_children)
        except KeyError:
            return set(), None, None, set()

    async def resolve_batch(self, queries, **kwargs):
        """
        A helper to execute a bunch of DNS requests.

        Args:
            queries (list): List of queries to resolve.
            **kwargs: Additional keyword arguments to pass to `resolve()`.

        Yields:
            tuple: A tuple containing the original query and its resolved value.

        Examples:
            >>> import asyncio
            >>> async def example_usage():
            ...     async for result in resolve_batch(['www.evilcorp.com', 'evilcorp.com']):
            ...         print(result)
            ('www.evilcorp.com', {'1.1.1.1'})
            ('evilcorp.com', {'2.2.2.2'})

        """
        for q in queries:
            yield (q, await self.resolve(q, **kwargs))

    def extract_targets(self, record):
        """
        Extracts hostnames or IP addresses from a given DNS record.

        This method reads the DNS record's type and based on that, extracts the target
        hostnames or IP addresses it points to. The type of DNS record
        (e.g., "A", "MX", "CNAME", etc.) determines which fields are used for extraction.

        Args:
            record (dns.rdata.Rdata): The DNS record to extract information from.

        Returns:
            set: A set of tuples, each containing the DNS record type and the extracted value.

        Examples:
            >>> from dns.rrset import from_text
            >>> record = from_text('www.example.com', 3600, 'IN', 'A', '192.0.2.1')
            >>> extract_targets(record[0])
            {('A', '192.0.2.1')}

            >>> record = from_text('example.com', 3600, 'IN', 'MX', '10 mail.example.com.')
            >>> extract_targets(record[0])
            {('MX', 'mail.example.com')}

        """
        results = set()
        rdtype = str(record.rdtype.name).upper()
        if rdtype in ("A", "AAAA", "NS", "CNAME", "PTR"):
            results.add((rdtype, self._clean_dns_record(record)))
        elif rdtype == "SOA":
            results.add((rdtype, self._clean_dns_record(record.mname)))
        elif rdtype == "MX":
            results.add((rdtype, self._clean_dns_record(record.exchange)))
        elif rdtype == "SRV":
            results.add((rdtype, self._clean_dns_record(record.target)))
        elif rdtype == "TXT":
            for s in record.strings:
                s = self.parent_helper.smart_decode(s)
                for match in dns_name_regex.finditer(s):
                    start, end = match.span()
                    host = s[start:end]
                    results.add((rdtype, host))
        elif rdtype == "NSEC":
            results.add((rdtype, self._clean_dns_record(record.next)))
        else:
            log.warning(f'Unknown DNS record type "{rdtype}"')
        return results

    @staticmethod
    def _clean_dns_record(record):
        """
        Cleans and formats a given DNS record for further processing.

        This static method converts the DNS record to text format if it's not already a string.
        It also removes any trailing dots and converts the record to lowercase.

        Args:
            record (str or dns.rdata.Rdata): The DNS record to clean.

        Returns:
            str: The cleaned and formatted DNS record.

        Examples:
            >>> _clean_dns_record('www.evilcorp.com.')
            'www.evilcorp.com'

            >>> from dns.rrset import from_text
            >>> record = from_text('www.evilcorp.com', 3600, 'IN', 'A', '1.2.3.4')[0]
            >>> _clean_dns_record(record)
            '1.2.3.4'
        """
        if not isinstance(record, str):
            record = str(record.to_text())
        return str(record).rstrip(".").lower()

    async def _catch(self, callback, *args, **kwargs):
        """
        Asynchronously catches exceptions thrown during DNS resolution and logs them.

        This method wraps around a given asynchronous callback function to handle different
        types of DNS exceptions and general exceptions. It logs the exceptions for debugging
        and, in some cases, re-raises them.

        Args:
            callback (callable): The asynchronous function to be executed.
            *args: Positional arguments to pass to the callback.
            **kwargs: Keyword arguments to pass to the callback.

        Returns:
            Any: The return value of the callback function, or an empty list if an exception is caught.

        Raises:
            dns.resolver.NoNameservers: When no nameservers could be reached.
        """
        try:
            return await callback(*args, **kwargs)
        except dns.resolver.NoNameservers:
            raise
        except (dns.exception.Timeout, dns.resolver.LifetimeTimeout, TimeoutError):
            log.debug(f"DNS query with args={args}, kwargs={kwargs} timed out after {self.timeout} seconds")
            raise
        except dns.exception.DNSException as e:
            self.debug(f"{e} (args={args}, kwargs={kwargs})")
        except Exception as e:
            log.warning(f"Error in {callback.__qualname__}() with args={args}, kwargs={kwargs}: {e}")
            log.trace(traceback.format_exc())
        return []

    async def is_wildcard(self, query, ips=None, rdtype=None):
        """
        Use this method to check whether a *host* is a wildcard entry

        This can reliably tell the difference between a valid DNS record and a wildcard within a wildcard domain.

        If you want to know whether a domain is using wildcard DNS, use `is_wildcard_domain()` instead.

        Args:
            query (str): The hostname to check for a wildcard entry.
            ips (list, optional): List of IPs to compare against, typically obtained from a previous DNS resolution of the query.
            rdtype (str, optional): The DNS record type (e.g., "A", "AAAA") to consider during the check.

        Returns:
            dict: A dictionary indicating if the query is a wildcard for each checked DNS record type.
                Keys are DNS record types like "A", "AAAA", etc.
                Values are tuples where the first element is a boolean indicating if the query is a wildcard,
                and the second element is the wildcard parent if it's a wildcard.

        Raises:
            ValueError: If only one of `ips` or `rdtype` is specified or if no valid IPs are specified.

        Examples:
            >>> is_wildcard("www.github.io")
            {"A": (True, "github.io"), "AAAA": (True, "github.io")}

            >>> is_wildcard("www.evilcorp.com", ips=["93.184.216.34"], rdtype="A")
            {"A": (False, "evilcorp.com")}

        Note:
            `is_wildcard` can be True, False, or None (indicating that wildcard detection was inconclusive)
        """
        result = {}

        if [ips, rdtype].count(None) == 1:
            raise ValueError("Both ips and rdtype must be specified")

        if not is_dns_name(query):
            return {}

        # skip check if the query's parent domain is excluded in the config
        for d in self.wildcard_ignore:
            if self.parent_helper.host_in_host(query, d):
                log.debug(f"Skipping wildcard detection on {query} because it is excluded in the config")
                return {}

        query = self._clean_dns_record(query)
        # skip check if it's an IP
        if is_ip(query) or not "." in query:
            return {}
        # skip check if the query is a domain
        if is_domain(query):
            return {}

        parent = parent_domain(query)
        parents = list(domain_parents(query))

        rdtypes_to_check = [rdtype] if rdtype is not None else self.all_rdtypes

        base_query_ips = dict()
        # if the caller hasn't already done the work of resolving the IPs
        if ips is None:
            # then resolve the query for all rdtypes
            for t in rdtypes_to_check:
                raw_results, errors = await self.resolve_raw(query, type=t, use_cache=True)
                if errors and not raw_results:
                    self.debug(f"Failed to resolve {query} ({t}) during wildcard detection")
                    result[t] = (None, parent)
                    continue
                for __rdtype, answers in raw_results:
                    base_query_results = set()
                    for answer in answers:
                        for _, t in self.extract_targets(answer):
                            base_query_results.add(t)
                    if base_query_results:
                        base_query_ips[__rdtype] = base_query_results
        else:
            # otherwise, we can skip all that
            cleaned_ips = set([self._clean_dns_record(ip) for ip in ips])
            if not cleaned_ips:
                raise ValueError("Valid IPs must be specified")
            base_query_ips[rdtype] = cleaned_ips
        if not base_query_ips:
            return result

        # once we've resolved the base query and have IP addresses to work with
        # we can compare the IPs to the ones we have on file for wildcards

        # for every parent domain, starting with the shortest
        try:
            for host in parents[::-1]:
                # make sure we've checked that domain for wildcards
                await self.is_wildcard_domain(host)

                # for every rdtype
                for _rdtype in list(base_query_ips):
                    # get the IPs from above
                    query_ips = base_query_ips.get(_rdtype, set())
                    host_hash = hash(host)

                    if host_hash in self._wildcard_cache:
                        # then get its IPs from our wildcard cache
                        wildcard_rdtypes = self._wildcard_cache[host_hash]

                        # then check to see if our IPs match the wildcard ones
                        if _rdtype in wildcard_rdtypes:
                            wildcard_ips = wildcard_rdtypes[_rdtype]
                            # if our IPs match the wildcard ones, then ladies and gentlemen we have a wildcard
                            is_wildcard = any(r in wildcard_ips for r in query_ips)

                            if is_wildcard and not result.get(_rdtype, (None, None))[0] is True:
                                result[_rdtype] = (True, host)

                    # if we've reached a point where the dns name is a complete wildcard, class can be dismissed early
                    base_query_rdtypes = set(base_query_ips)
                    wildcard_rdtypes_set = set([k for k, v in result.items() if v[0] is True])
                    if base_query_rdtypes and wildcard_rdtypes_set and base_query_rdtypes == wildcard_rdtypes_set:
                        log.debug(
                            f"Breaking from wildcard detection for {query} at {host} because base query rdtypes ({base_query_rdtypes}) == wildcard rdtypes ({wildcard_rdtypes_set})"
                        )
                        raise DNSWildcardBreak()
        except DNSWildcardBreak:
            pass

        return result

    async def is_wildcard_domain(self, domain, log_info=False):
        """
        Check whether a given host or its children make use of wildcard DNS entries. Wildcard DNS can have
        various implications, particularly in subdomain enumeration and subdomain takeovers.

        Args:
            domain (str): The domain to check for wildcard DNS entries.
            log_info (bool, optional): Whether to log the result of the check. Defaults to False.

        Returns:
            dict: A dictionary where the keys are the parent domains that have wildcard DNS entries,
            and the values are another dictionary of DNS record types ("A", "AAAA", etc.) mapped to
            sets of their resolved IP addresses.

        Examples:
            >>> is_wildcard_domain("github.io")
            {"github.io": {"A": {"1.2.3.4"}, "AAAA": {"dead::beef"}}}

            >>> is_wildcard_domain("example.com")
            {}
        """
        wildcard_domain_results = {}
        domain = self._clean_dns_record(domain)

        if not is_dns_name(domain):
            return {}

        # skip check if the query's parent domain is excluded in the config
        for d in self.wildcard_ignore:
            if self.parent_helper.host_in_host(domain, d):
                log.debug(f"Skipping wildcard detection on {domain} because it is excluded in the config")
                return {}

        rdtypes_to_check = set(self.all_rdtypes)

        # make a list of its parents
        parents = list(domain_parents(domain, include_self=True))
        # and check each of them, beginning with the highest parent (i.e. the root domain)
        for i, host in enumerate(parents[::-1]):
            # have we checked this host before?
            host_hash = hash(host)
            async with self._wildcard_lock.lock(host_hash):
                # if we've seen this host before
                if host_hash in self._wildcard_cache:
                    wildcard_domain_results[host] = self._wildcard_cache[host_hash]
                    continue

                log.verbose(f"Checking if {host} is a wildcard")

                # determine if this is a wildcard domain

                # resolve a bunch of random subdomains of the same parent
                is_wildcard = False
                wildcard_results = dict()
                for rdtype in list(rdtypes_to_check):
                    # continue if a wildcard was already found for this rdtype
                    # if rdtype in self._wildcard_cache[host_hash]:
                    #     continue
                    for _ in range(self.wildcard_tests):
                        rand_query = f"{rand_string(digits=False, length=10)}.{host}"
                        results = await self.resolve(rand_query, type=rdtype, use_cache=False)
                        if results:
                            is_wildcard = True
                            if not rdtype in wildcard_results:
                                wildcard_results[rdtype] = set()
                            wildcard_results[rdtype].update(results)
                            # we know this rdtype is a wildcard
                            # so we don't need to check it anymore
                            with suppress(KeyError):
                                rdtypes_to_check.remove(rdtype)

                self._wildcard_cache.update({host_hash: wildcard_results})
                wildcard_domain_results.update({host: wildcard_results})
                if is_wildcard:
                    wildcard_rdtypes_str = ",".join(sorted([t.upper() for t, r in wildcard_results.items() if r]))
                    log_fn = log.verbose
                    if log_info:
                        log_fn = log.info
                    log_fn(f"Encountered domain with wildcard DNS ({wildcard_rdtypes_str}): {host}")
                else:
                    log.verbose(f"Finished checking {host}, it is not a wildcard")

        return wildcard_domain_results

    async def _connectivity_check(self, interval=5):
        """
        Periodically checks for an active internet connection by attempting DNS resolution.

        Args:
            interval (int, optional): The time interval, in seconds, at which to perform the check.
            Defaults to 5 seconds.

        Returns:
            bool: True if there is an active internet connection, False otherwise.

        Examples:
            >>> await _connectivity_check()
            True
        """
        if self._last_dns_success is not None:
            if time.time() - self._last_dns_success < interval:
                return True
        dns_server_working = []
        async with self._dns_connectivity_lock:
            with suppress(Exception):
                dns_server_working = await self._catch(self.resolver.resolve, "www.google.com", rdtype="A")
                if dns_server_working:
                    self._last_dns_success = time.time()
                    return True
        if time.time() - self._last_connectivity_warning > interval:
            log.warning(f"DNS queries are failing, please check your internet connection")
            self._last_connectivity_warning = time.time()
        self._errors.clear()
        return False

    def _parse_rdtype(self, t, default=None):
        if isinstance(t, str):
            if t.strip().lower() in ("any", "all", "*"):
                return self.all_rdtypes
            else:
                return [t.strip().upper()]
        elif any([isinstance(t, x) for x in (list, tuple)]):
            return [str(_).strip().upper() for _ in t]
        return default

    def debug(self, *args, **kwargs):
        if self._debug:
            log.trace(*args, **kwargs)

    def _get_dummy_module(self, name):
        try:
            dummy_module = self._dummy_modules[name]
        except KeyError:
            dummy_module = self.parent_helper._make_dummy_module(name=name, _type="DNS")
            dummy_module.suppress_dupes = False
            self._dummy_modules[name] = dummy_module
        return dummy_module



from ..bbot_fixtures import *  # noqa: F401


@pytest.mark.asyncio
async def test_cloud_helpers(bbot_scanner, bbot_config):
    scan1 = bbot_scanner("127.0.0.1", config=bbot_config)

    provider_names = ("amazon", "google", "azure", "digitalocean", "oracle", "akamai", "cloudflare", "github")
    for provider_name in provider_names:
        assert provider_name in scan1.helpers.cloud.providers.providers

    for p in scan1.helpers.cloud.providers.providers.values():
        print(f"{p.name}: {p.domains} / {p.ranges}")
    amazon_ranges = list(scan1.helpers.cloud["amazon"].ranges)
    assert amazon_ranges
    amazon_range = next(iter(amazon_ranges))
    amazon_address = amazon_range.broadcast_address

    ip_event = scan1.make_event(amazon_address, source=scan1.root_event)
    aws_event1 = scan1.make_event("amazonaws.com", source=scan1.root_event)
    aws_event2 = scan1.make_event("asdf.amazonaws.com", source=scan1.root_event)
    aws_event3 = scan1.make_event("asdfamazonaws.com", source=scan1.root_event)
    aws_event4 = scan1.make_event("test.asdf.aws", source=scan1.root_event)

    other_event1 = scan1.make_event("cname.evilcorp.com", source=scan1.root_event)
    other_event2 = scan1.make_event("cname2.evilcorp.com", source=scan1.root_event)
    other_event3 = scan1.make_event("cname3.evilcorp.com", source=scan1.root_event)
    other_event2._resolved_hosts = {amazon_address}
    other_event3._resolved_hosts = {"asdf.amazonaws.com"}

    for event in (ip_event, aws_event1, aws_event2, aws_event4, other_event2, other_event3):
        await scan1.helpers.cloud.tag_event(event)
        assert "cloud-amazon" in event.tags, f"{event} was not properly cloud-tagged"

    for event in (aws_event3, other_event1):
        await scan1.helpers.cloud.tag_event(event)
        assert "cloud-amazon" not in event.tags, f"{event} was improperly cloud-tagged"
        assert not any(
            t for t in event.tags if t.startswith("cloud-") or t.startswith("cdn-")
        ), f"{event} was improperly cloud-tagged"

    google_event1 = scan1.make_event("asdf.googleapis.com", source=scan1.root_event)
    google_event2 = scan1.make_event("asdf.google", source=scan1.root_event)
    google_event3 = scan1.make_event("asdf.evilcorp.com", source=scan1.root_event)
    google_event3._resolved_hosts = {"asdf.storage.googleapis.com"}

    for event in (google_event1, google_event2, google_event3):
        await scan1.helpers.cloud.tag_event(event)
        assert "cloud-google" in event.tags, f"{event} was not properly cloud-tagged"
    assert "cloud-storage-bucket" in google_event3.tags


@pytest.mark.asyncio
async def test_cloud_helpers_excavate(bbot_scanner, bbot_config, bbot_httpserver):
    url = bbot_httpserver.url_for("/test_cloud_helpers_excavate")
    bbot_httpserver.expect_request(uri="/test_cloud_helpers_excavate").respond_with_data(
        "<a href='asdf.s3.amazonaws.com'/>"
    )
    scan1 = bbot_scanner(url, modules=["httpx", "excavate"], config=bbot_config)
    events = [e async for e in scan1.async_start()]
    assert 1 == len(
        [
            e
            for e in events
            if e.type == "STORAGE_BUCKET"
            and e.data["name"] == "asdf"
            and "cloud-amazon" in e.tags
            and "cloud-storage-bucket" in e.tags
        ]
    )


@pytest.mark.asyncio
async def test_cloud_helpers_speculate(bbot_scanner, bbot_config):
    scan1 = bbot_scanner("asdf.s3.amazonaws.com", modules=["speculate"], config=bbot_config)
    events = [e async for e in scan1.async_start()]
    assert 1 == len(
        [
            e
            for e in events
            if e.type == "STORAGE_BUCKET"
            and e.data["name"] == "asdf"
            and "cloud-amazon" in e.tags
            and "cloud-storage-bucket" in e.tags
        ]
    )



import json
from contextlib import suppress

from bbot.modules.templates.portscanner import portscanner


class masscan(portscanner):
    flags = ["active", "portscan", "aggressive"]
    watched_events = ["IP_ADDRESS", "IP_RANGE"]
    produced_events = ["OPEN_TCP_PORT"]
    meta = {"description": "Port scan with masscan. By default, scans top 100 ports."}
    options = {
        "top_ports": 100,
        "ports": "",
        # ping scan at 600 packets/s ~= entire private IP space in 8 hours
        "rate": 600,
        "wait": 5,
        "ping_first": False,
        "ping_only": False,
        "use_cache": False,
    }
    options_desc = {
        "top_ports": "Top ports to scan (default 100) (to override, specify 'ports')",
        "ports": "Ports to scan",
        "rate": "Rate in packets per second",
        "wait": "Seconds to wait for replies after scan is complete",
        "ping_first": "Only portscan hosts that reply to pings",
        "ping_only": "Ping sweep only, no portscan",
        "use_cache": "Instead of scanning, use the results from the previous scan",
    }
    deps_ansible = [
        {
            "name": "install dev tools",
            "package": {"name": ["gcc", "git", "make"], "state": "present"},
            "become": True,
            "ignore_errors": True,
        },
        {
            "name": "Download masscan source code",
            "git": {
                "repo": "https://github.com/robertdavidgraham/masscan.git",
                "dest": "#{BBOT_TEMP}/masscan",
                "single_branch": True,
                "version": "master",
            },
        },
        {
            "name": "Build masscan",
            "command": {
                "chdir": "#{BBOT_TEMP}/masscan",
                "cmd": "make -j",
                "creates": "#{BBOT_TEMP}/masscan/bin/masscan",
            },
        },
        {
            "name": "Install masscan",
            "copy": {"src": "#{BBOT_TEMP}/masscan/bin/masscan", "dest": "#{BBOT_TOOLS}/", "mode": "u+x,g+x,o+x"},
        },
    ]
    batch_size = 1000000

    async def setup(self):
        self.top_ports = self.config.get("top_ports", 100)
        self.rate = self.config.get("rate", 600)
        self.wait = self.config.get("wait", 10)
        self.ping_first = self.config.get("ping_first", False)
        self.ping_only = self.config.get("ping_only", False)
        self.use_cache = self.config.get("use_cache", False)
        self.ports = self.config.get("ports", "")
        if self.ports:
            try:
                self.helpers.parse_port_string(self.ports)
            except ValueError as e:
                return False, f"Error parsing ports: {e}"
        self.alive_hosts = dict()

        _, invalid_targets = self._build_targets(self.scan.target)
        if invalid_targets > 0:
            self.warning(
                f"Masscan can only accept IP addresses or IP ranges as target ({invalid_targets:,} targets were hostnames)"
            )

        self.run_time = self.helpers.make_date()
        self.ping_cache = self.scan.home / f"masscan_ping.txt"
        self.syn_cache = self.scan.home / f"masscan_syn.txt"
        if self.use_cache:
            files_exist = self.ping_cache.is_file() or self.syn_cache.is_file()
            files_empty = self.helpers.filesize(self.ping_cache) == 0 and self.helpers.filesize(self.syn_cache) == 0
            if not files_exist:
                return (
                    False,
                    f"use_cache is True but could not find cache file at {self.ping_cache} or {self.syn_cache}",
                )
            if files_empty:
                return (
                    False,
                    f"use_cache is True but could cached files {self.ping_cache} and {self.syn_cache} are empty",
                )
        else:
            self.helpers.depsinstaller.ensure_root(message="Masscan requires root privileges")
        self.ping_cache_fd = None
        self.syn_cache_fd = None

        return await super().setup()

    async def handle_batch(self, *events):
        if self.use_cache:
            await self.emit_from_cache()
        else:
            targets = [str(e.data) for e in events]
            if not targets:
                self.warning("No targets specified")
                return

            # ping scan
            if self.ping_first or self.ping_only:
                self.verbose("Starting masscan (ping scan)")

                await self.masscan(targets, result_callback=self.append_alive_host, ping=True)
                targets = ",".join(str(h) for h in self.alive_hosts)
                if not targets:
                    self.warning("No hosts responded to pings")
                    return

            # TCP SYN scan
            if not self.ping_only:
                self.verbose("Starting masscan (TCP SYN scan)")
                await self.masscan(targets, result_callback=self.emit_open_tcp_port)
            else:
                self.verbose("Only ping sweep was requested, skipping TCP SYN scan")
            # save memory
            self.alive_hosts.clear()

    async def masscan(self, targets, result_callback, ping=False):
        target_file = self.helpers.tempfile(targets, pipe=False)
        command = self._build_masscan_command(target_file, ping=ping)
        stats_file = self.helpers.tempfile_tail(callback=self.verbose)
        try:
            with open(stats_file, "w") as stats_fh:
                async for line in self.helpers.run_live(command, sudo=True, stderr=stats_fh):
                    await self.process_output(line, result_callback=result_callback)
        finally:
            for file in (stats_file, target_file):
                file.unlink()

    def _build_masscan_command(self, target_file=None, dry_run=False, ping=False):
        command = (
            "masscan",
            "--excludefile",
            str(self.exclude_file),
            "--rate",
            self.rate,
            "--wait",
            self.wait,
            "--open-only",
            "-oJ",
            "-",
        )
        if target_file is not None:
            command += ("-iL", str(target_file))
        if ping:
            command += ("--ping",)
        else:
            if self.ports:
                command += ("-p", self.ports)
            else:
                command += ("--top-ports", str(self.top_ports))
        if dry_run:
            command += ("--echo",)
        return command

    async def process_output(self, line, result_callback):
        try:
            j = json.loads(line)
        except Exception:
            return
        ip = j.get("ip", "")
        if not ip:
            return
        ports = j.get("ports", [])
        if not ports:
            return
        for p in ports:
            proto = p.get("proto", "")
            port_number = p.get("port", "")
            if proto == "" or port_number == "":
                continue
            result = str(ip)
            source = None
            with suppress(KeyError):
                source = self.alive_hosts[ip]
            if proto != "icmp":
                result = self.helpers.make_netloc(result, port_number)
                if source is None:
                    source = self.make_event(ip, "IP_ADDRESS", source=self.get_source_event(ip))
                    await self.emit_event(source)
            await result_callback(result, source=source)

    async def append_alive_host(self, host, source):
        host_event = self.make_event(host, "IP_ADDRESS", source=self.get_source_event(host))
        self.alive_hosts[host] = host_event
        self._write_ping_result(host)
        await self.emit_event(host_event)

    async def emit_open_tcp_port(self, data, source):
        self._write_syn_result(data)
        await self.emit_event(data, "OPEN_TCP_PORT", source=source)

    async def emit_from_cache(self):
        ip_events = {}
        # ping scan
        if self.ping_cache.is_file():
            cached_pings = list(self.helpers.read_file(self.ping_cache))
            if cached_pings:
                self.success(f"{len(cached_pings):,} hosts loaded from previous ping scan")
            else:
                self.verbose(f"No hosts cached from previous ping scan")
            for ip in cached_pings:
                if self.scan.stopping:
                    break
                ip_event = self.make_event(ip, "IP_ADDRESS", source=self.get_source_event(ip))
                ip_events[ip] = ip_event
                await self.emit_event(ip_event)
        # syn scan
        if self.syn_cache.is_file():
            cached_syns = list(self.helpers.read_file(self.syn_cache))
            if cached_syns:
                self.success(f"{len(cached_syns):,} hosts loaded from previous SYN scan")
            else:
                self.warning(f"No hosts cached from previous SYN scan")
            for line in cached_syns:
                if self.scan.stopping:
                    break
                host, port = self.helpers.split_host_port(line)
                host = str(host)
                source_event = ip_events.get(host)
                if source_event is None:
                    self.verbose(f"Source event not found for {line}")
                    source_event = self.make_event(line, "IP_ADDRESS", source=self.get_source_event(line))
                    await self.emit_event(source_event)
                await self.emit_event(line, "OPEN_TCP_PORT", source=source_event)

    def get_source_event(self, host):
        source_event = self.scan.target.get(host)
        if source_event is None:
            source_event = self.scan.whitelist.get(host)
            if source_event is None:
                source_event = self.scan.root_event
        return source_event

    async def cleanup(self):
        if self.ping_first:
            with suppress(Exception):
                self.ping_cache_fd.close()
        with suppress(Exception):
            self.syn_cache_fd.close()
        with suppress(Exception):
            self.exclude_file.unlink()

    def _write_ping_result(self, host):
        if self.ping_cache_fd is None:
            self.helpers.backup_file(self.ping_cache)
            self.ping_cache_fd = open(self.ping_cache, "w")
        self.ping_cache_fd.write(f"{host}\n")
        self.ping_cache_fd.flush()

    def _write_syn_result(self, data):
        if self.syn_cache_fd is None:
            self.helpers.backup_file(self.syn_cache)
            self.syn_cache_fd = open(self.syn_cache, "w")
        self.syn_cache_fd.write(f"{data}\n")
        self.syn_cache_fd.flush()



import os
import dns
import sys
import pytest
import asyncio  # noqa
import logging
import subprocess
import tldextract
import pytest_httpserver
from pathlib import Path
from omegaconf import OmegaConf

from werkzeug.wrappers import Request


class SubstringRequestMatcher(pytest_httpserver.httpserver.RequestMatcher):
    def match_data(self, request: Request) -> bool:
        if self.data is None:
            return True
        return self.data in request.data


pytest_httpserver.httpserver.RequestMatcher = SubstringRequestMatcher


test_config = OmegaConf.load(Path(__file__).parent / "test.conf")
if test_config.get("debug", False):
    os.environ["BBOT_DEBUG"] = "True"

from .bbot_fixtures import *  # noqa: F401
import bbot.core.logger  # noqa: F401
from bbot.core.errors import *  # noqa: F401

# silence pytest_httpserver
log = logging.getLogger("werkzeug")
log.setLevel(logging.CRITICAL)

# silence stdout
root_logger = logging.getLogger()
for h in root_logger.handlers:
    h.addFilter(lambda x: x.levelname not in ("STDOUT", "TRACE"))

tldextract.extract("www.evilcorp.com")

log = logging.getLogger(f"bbot.test.fixtures")


@pytest.fixture
def bbot_scanner():
    from bbot.scanner import Scanner

    return Scanner


@pytest.fixture
def neograph(monkeypatch, helpers):
    helpers.depsinstaller.pip_install(["py2neo"])

    class NeoGraph:
        def __init__(self, *args, **kwargs):
            pass

        def merge(self, *args, **kwargs):
            return True

    import py2neo

    monkeypatch.setattr(py2neo, "Graph", NeoGraph)
    from bbot.db.neo4j import Neo4j

    return Neo4j(uri="bolt://127.0.0.1:1111")


@pytest.fixture
def scan(monkeypatch, bbot_config):
    from bbot.scanner import Scanner

    bbot_scan = Scanner("127.0.0.1", modules=["ipneighbor"], config=bbot_config)

    fallback_nameservers_file = bbot_scan.helpers.bbot_home / "fallback_nameservers.txt"
    with open(fallback_nameservers_file, "w") as f:
        f.write("8.8.8.8\n")
    monkeypatch.setattr(bbot_scan.helpers.dns, "fallback_nameservers_file", fallback_nameservers_file)

    return bbot_scan


@pytest.fixture
def helpers(scan):
    return scan.helpers


httpx_response = {
    "timestamp": "2022-11-14T12:14:27.377566416-05:00",
    "hash": {
        "body_md5": "84238dfc8092e5d9c0dac8ef93371a07",
        "body_mmh3": "-1139337416",
        "body_sha256": "ea8fac7c65fb589b0d53560f5251f74f9e9b243478dcb6b3ea79b5e36449c8d9",
        "body_simhash": "9899951357530060719",
        "header_md5": "6e483c85c3b9b96f0e33d84237ca651e",
        "header_mmh3": "-957156428",
        "header_sha256": "5a809d8a53aded843179237365bb6dd069fba75ff8603ac2f6dc6c05d6bf0e76",
        "header_simhash": "15614709017155972941",
    },
    "port": "80",
    "url": "http://example.com:80",
    "input": "http://example.com:80",
    "title": "Example Domain",
    "scheme": "http",
    "webserver": "ECS (agb/A445)",
    "body": '<!doctype html>\n<html>\n<head>\n    <title>Example Domain</title>\n\n    <meta charset="utf-8" />\n    <meta http-equiv="Content-type" content="text/html; charset=utf-8" />\n    <meta name="viewport" content="width=device-width, initial-scale=1" />\n    <style type="text/css">\n    body {\n        background-color: #f0f0f2;\n        margin: 0;\n        padding: 0;\n        font-family: -apple-system, system-ui, BlinkMacSystemFont, "Segoe UI", "Open Sans", "Helvetica Neue", Helvetica, Arial, sans-serif;\n        \n    }\n    div {\n        width: 600px;\n        margin: 5em auto;\n        padding: 2em;\n        background-color: #fdfdff;\n        border-radius: 0.5em;\n        box-shadow: 2px 3px 7px 2px rgba(0,0,0,0.02);\n    }\n    a:link, a:visited {\n        color: #38488f;\n        text-decoration: none;\n    }\n    @media (max-width: 700px) {\n        div {\n            margin: 0 auto;\n            width: auto;\n        }\n    }\n    </style>    \n</head>\n\n<body>\n<div>\n    <h1>Example Domain</h1>\n    <p>This domain is for use in illustrative examples in documents. You may use this\n    domain in literature without prior coordination or asking for permission.</p>\n    <p><a href="https://www.iana.org/domains/example">More information...</a></p>\n</div>\n</body>\n</html>\n',
    "content_type": "text/html",
    "method": "GET",
    "host": "93.184.216.34",
    "path": "/",
    "header": {
        "age": "526111",
        "cache_control": "max-age=604800",
        "content_type": "text/html; charset=UTF-8",
        "date": "Mon, 14 Nov 2022 17:14:27 GMT",
        "etag": '"3147526947+ident+gzip"',
        "expires": "Mon, 21 Nov 2022 17:14:27 GMT",
        "last_modified": "Thu, 17 Oct 2019 07:18:26 GMT",
        "server": "ECS (agb/A445)",
        "vary": "Accept-Encoding",
        "x_cache": "HIT",
    },
    "raw_header": 'HTTP/1.1 200 OK\r\nConnection: close\r\nAge: 526111\r\nCache-Control: max-age=604800\r\nContent-Type: text/html; charset=UTF-8\r\nDate: Mon, 14 Nov 2022 17:14:27 GMT\r\nEtag: "3147526947+ident+gzip"\r\nExpires: Mon, 21 Nov 2022 17:14:27 GMT\r\nLast-Modified: Thu, 17 Oct 2019 07:18:26 GMT\r\nServer: ECS (agb/A445)\r\nVary: Accept-Encoding\r\nX-Cache: HIT\r\n\r\n',
    "request": "GET / HTTP/1.1\r\nHost: example.com\r\nUser-Agent: Mozilla/5.0 (SymbianOS/9.1; U; de) AppleWebKit/413 (KHTML, like Gecko) Safari/413\r\nAccept-Charset: utf-8\r\nAccept-Encoding: gzip\r\n\r\n",
    "time": "112.128324ms",
    "a": ["93.184.216.34", "2606:2800:220:1:248:1893:25c8:1946"],
    "words": 298,
    "lines": 47,
    "status_code": 200,
    "content_length": 1256,
    "failed": False,
}


@pytest.fixture
def events(scan):
    class bbot_events:
        localhost = scan.make_event("127.0.0.1", source=scan.root_event)
        ipv4 = scan.make_event("8.8.8.8", source=scan.root_event)
        netv4 = scan.make_event("8.8.8.8/30", source=scan.root_event)
        ipv6 = scan.make_event("2001:4860:4860::8888", source=scan.root_event)
        netv6 = scan.make_event("2001:4860:4860::8888/126", source=scan.root_event)
        domain = scan.make_event("publicAPIs.org", source=scan.root_event)
        subdomain = scan.make_event("api.publicAPIs.org", source=scan.root_event)
        email = scan.make_event("bob@evilcorp.co.uk", "EMAIL_ADDRESS", source=scan.root_event)
        open_port = scan.make_event("api.publicAPIs.org:443", source=scan.root_event)
        protocol = scan.make_event(
            {"host": "api.publicAPIs.org", "port": 443, "protocol": "HTTP"}, "PROTOCOL", source=scan.root_event
        )
        ipv4_open_port = scan.make_event("8.8.8.8:443", source=scan.root_event)
        ipv6_open_port = scan.make_event("[2001:4860:4860::8888]:443", "OPEN_TCP_PORT", source=scan.root_event)
        url_unverified = scan.make_event("https://api.publicAPIs.org:443/hellofriend", source=scan.root_event)
        ipv4_url_unverified = scan.make_event("https://8.8.8.8:443/hellofriend", source=scan.root_event)
        ipv6_url_unverified = scan.make_event("https://[2001:4860:4860::8888]:443/hellofriend", source=scan.root_event)
        url = scan.make_event(
            "https://api.publicAPIs.org:443/hellofriend", "URL", tags=["status-200"], source=scan.root_event
        )
        ipv4_url = scan.make_event(
            "https://8.8.8.8:443/hellofriend", "URL", tags=["status-200"], source=scan.root_event
        )
        ipv6_url = scan.make_event(
            "https://[2001:4860:4860::8888]:443/hellofriend", "URL", tags=["status-200"], source=scan.root_event
        )
        url_hint = scan.make_event("https://api.publicAPIs.org:443/hello.ash", "URL_HINT", source=url)
        vulnerability = scan.make_event(
            {"host": "evilcorp.com", "severity": "INFO", "description": "asdf"},
            "VULNERABILITY",
            source=scan.root_event,
        )
        finding = scan.make_event({"host": "evilcorp.com", "description": "asdf"}, "FINDING", source=scan.root_event)
        vhost = scan.make_event({"host": "evilcorp.com", "vhost": "www.evilcorp.com"}, "VHOST", source=scan.root_event)
        http_response = scan.make_event(httpx_response, "HTTP_RESPONSE", source=scan.root_event)
        storage_bucket = scan.make_event(
            {"name": "storage", "url": "https://storage.blob.core.windows.net"},
            "STORAGE_BUCKET",
            source=scan.root_event,
        )
        emoji = scan.make_event("💩", "WHERE_IS_YOUR_GOD_NOW", source=scan.root_event)

    bbot_events.all = [  # noqa: F841
        bbot_events.localhost,
        bbot_events.ipv4,
        bbot_events.netv4,
        bbot_events.ipv6,
        bbot_events.netv6,
        bbot_events.domain,
        bbot_events.subdomain,
        bbot_events.email,
        bbot_events.open_port,
        bbot_events.protocol,
        bbot_events.ipv4_open_port,
        bbot_events.ipv6_open_port,
        bbot_events.url_unverified,
        bbot_events.ipv4_url_unverified,
        bbot_events.ipv6_url_unverified,
        bbot_events.url,
        bbot_events.ipv4_url,
        bbot_events.ipv6_url,
        bbot_events.url_hint,
        bbot_events.vulnerability,
        bbot_events.finding,
        bbot_events.vhost,
        bbot_events.http_response,
        bbot_events.storage_bucket,
        bbot_events.emoji,
    ]

    for e in bbot_events.all:
        e.scope_distance = 0

    return bbot_events


@pytest.fixture
def agent(monkeypatch, bbot_config):
    from bbot import agent

    test_agent = agent.Agent(bbot_config)
    test_agent.setup()
    return test_agent


# bbot config
from bbot import config as default_config

test_config = OmegaConf.load(Path(__file__).parent / "test.conf")
test_config = OmegaConf.merge(default_config, test_config)

if test_config.get("debug", False):
    logging.getLogger("bbot").setLevel(logging.DEBUG)


@pytest.fixture
def bbot_config():
    return test_config


from bbot.modules import module_loader

available_modules = list(module_loader.configs(type="scan"))
available_output_modules = list(module_loader.configs(type="output"))
available_internal_modules = list(module_loader.configs(type="internal"))


@pytest.fixture(autouse=True)
def install_all_python_deps():
    deps_pip = set()
    for module in module_loader.preloaded().values():
        deps_pip.update(set(module.get("deps", {}).get("pip", [])))
    subprocess.run([sys.executable, "-m", "pip", "install"] + list(deps_pip))


class MockResolver:
    import dns

    def __init__(self, mock_data=None):
        self.mock_data = mock_data if mock_data else {}
        self.nameservers = ["127.0.0.1"]

    async def resolve_address(self, ipaddr, *args, **kwargs):
        modified_kwargs = {}
        modified_kwargs.update(kwargs)
        modified_kwargs["rdtype"] = "PTR"
        return await self.resolve(str(dns.reversename.from_address(ipaddr)), *args, **modified_kwargs)

    def create_dns_response(self, query_name, rdtype):
        query_name = query_name.strip(".")
        answers = self.mock_data.get(query_name, {}).get(rdtype, [])
        if not answers:
            raise self.dns.resolver.NXDOMAIN(f"No answer found for {query_name} {rdtype}")

        message_text = f"""id 1234
opcode QUERY
rcode NOERROR
flags QR AA RD
;QUESTION
{query_name}. IN {rdtype}
;ANSWER"""
        for answer in answers:
            message_text += f"\n{query_name}. 1 IN {rdtype} {answer}"

        message_text += "\n;AUTHORITY\n;ADDITIONAL\n"
        message = self.dns.message.from_text(message_text)
        return message

    async def resolve(self, query_name, rdtype=None):
        if rdtype is None:
            rdtype = "A"
        elif isinstance(rdtype, str):
            rdtype = rdtype.upper()
        else:
            rdtype = str(rdtype.name).upper()

        domain_name = self.dns.name.from_text(query_name)
        rdtype_obj = self.dns.rdatatype.from_text(rdtype)

        if "_NXDOMAIN" in self.mock_data and query_name in self.mock_data["_NXDOMAIN"]:
            # Simulate the NXDOMAIN exception
            raise self.dns.resolver.NXDOMAIN

        try:
            response = self.create_dns_response(query_name, rdtype)
            answer = self.dns.resolver.Answer(domain_name, rdtype_obj, self.dns.rdataclass.IN, response)
            return answer
        except self.dns.resolver.NXDOMAIN:
            return []


@pytest.fixture()
def mock_dns():
    def _mock_dns(scan, mock_data):
        scan.helpers.dns.resolver = MockResolver(mock_data)

    return _mock_dns



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class hunterio(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["EMAIL_ADDRESS", "DNS_NAME", "URL_UNVERIFIED"]
    flags = ["passive", "email-enum", "subdomain-enum", "safe"]
    meta = {"description": "Query hunter.io for emails", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "Hunter.IO API key"}

    base_url = "https://api.hunter.io/v2"
    limit = 100

    async def ping(self):
        url = f"{self.base_url}/account?api_key={self.api_key}"
        r = await self.helpers.request(url)
        resp_content = getattr(r, "text", "")
        assert getattr(r, "status_code", 0) == 200, resp_content

    async def handle_event(self, event):
        query = self.make_query(event)
        for entry in await self.query(query):
            email = entry.get("value", "")
            sources = entry.get("sources", [])
            if email:
                email_event = self.make_event(email, "EMAIL_ADDRESS", event)
                if email_event:
                    await self.emit_event(email_event)
                    for source in sources:
                        domain = source.get("domain", "")
                        if domain:
                            await self.emit_event(domain, "DNS_NAME", email_event)
                        url = source.get("uri", "")
                        if url:
                            await self.emit_event(url, "URL_UNVERIFIED", email_event)

    async def query(self, query):
        emails = []
        url = (
            f"{self.base_url}/domain-search?domain={query}&api_key={self.api_key}"
            + "&limit={page_size}&offset={offset}"
        )
        agen = self.helpers.api_page_iter(url, page_size=self.limit)
        try:
            async for j in agen:
                new_emails = j.get("data", {}).get("emails", [])
                if not new_emails:
                    break
                emails += new_emails
        finally:
            agen.aclose()
        return emails



from .base import ModuleTestBase


class TestAzure_Realm(ModuleTestBase):
    targets = ["evilcorp.com"]
    config_overrides = {"scope_report_distance": 1}

    response_json = {
        "State": 3,
        "UserState": 2,
        "Login": "test@evilcorp.com",
        "NameSpaceType": "Federated",
        "DomainName": "evilcorp.com",
        "FederationGlobalVersion": -1,
        "AuthURL": "https://evilcorp.okta.com/app/office365/deadbeef/sso/wsfed/passive?username=test%40evilcorp.com&wa=wsignin1.0&wtrevilcorplm=urn%3afederation%3aMicrosoftOnline&wctx=",
        "FederationBrandName": "EvilCorp",
        "AuthNForwardType": 1,
        "CloudInstanceName": "microsoftonline.com",
        "CloudInstanceIssuerUri": "urn:federation:MicrosoftOnline",
    }

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url=f"https://login.microsoftonline.com/getuserrealm.srf?login=test@evilcorp.com",
            json=self.response_json,
        )

    def check(self, module_test, events):
        assert any(
            e.data == "https://evilcorp.okta.com/app/office365/deadbeef/sso/wsfed/passive" for e in events
        ), "Failed to detect URL"



from bbot.modules.templates.subdomain_enum import subdomain_enum


class shodan(subdomain_enum):
    options = {"api_key": ""}
    options_desc = {"api_key": "Shodan API key"}

    base_url = "https://api.shodan.io"

    async def setup(self):
        await super().setup()
        self.api_key = None
        for module_name in ("shodan", "shodan_dns", "shodan_port"):
            module_config = self.scan.config.get("modules", {}).get(module_name, {})
            api_key = module_config.get("api_key", "")
            if api_key:
                self.api_key = api_key
                break
        if not self.api_key:
            if self.auth_required:
                return None, "No API key set"
        try:
            await self.ping()
            self.hugesuccess(f"API is ready")
            return True
        except Exception as e:
            return None, f"Error with API ({str(e).strip()})"
        return True

    async def ping(self):
        url = f"{self.base_url}/api-info?key={self.api_key}"
        r = await self.request_with_fail_count(url)
        resp_content = getattr(r, "text", "")
        assert getattr(r, "status_code", 0) == 200, resp_content



from bbot.modules.base import BaseModule

# the following are the result of a 1-day internet survey to find the top SRV records
# the scan resulted in 36,282 SRV records. the count for each one is shown.
common_srvs = [
    "_sipfederationtls._tcp",  # 6909
    "_sip._tls",  # 6853
    "_autodiscover._tcp",  # 4268
    "_xmpp-server._tcp",  # 1437
    "_sip._tcp",  # 1193
    "_sips._tcp",  # 1183
    "_caldavs._tcp",  # 1179
    "_carddavs._tcp",  # 1132
    "_caldav._tcp",  # 1035
    "_carddav._tcp",  # 1024
    "_sip._udp",  # 1007
    "_imaps._tcp",  # 1007
    "_submission._tcp",  # 906
    "_h323cs._tcp",  # 846
    "_h323ls._udp",  # 782
    "_xmpp-client._tcp",  # 689
    "_pop3s._tcp",  # 394
    "_jabber._tcp",  # 277
    "_imap._tcp",  # 267
    "_turn._udp",  # 256
    "_pop3._tcp",  # 221
    "_ldap._tcp",  # 213
    "_smtps._tcp",  # 195
    "_sipinternaltls._tcp",  # 192
    "_vlmcs._tcp",  # 165
    "_kerberos._udp",  # 163
    "_kerberos._tcp",  # 148
    "_kpasswd._udp",  # 128
    "_kpasswd._tcp",  # 100
    "_ntp._udp",  # 90
    "_gc._tcp",  # 73
    "_kerberos-master._udp",  # 66
    "_ldap._tcp.dc._msdcs",  # 63
    "_matrix._tcp",  # 62
    "_smtp._tcp",  # 61
    "_stun._udp",  # 57
    "_kerberos._tcp.dc._msdcs",  # 54
    "_ldap._tcp.gc._msdcs",  # 49
    "_kerberos-adm._tcp",  # 44
    "_ldap._tcp.pdc._msdcs",  # 43
    "_kerberos-master._tcp",  # 43
    "_http._tcp",  # 37
    "_h323rs._tcp",  # 36
    "_sipinternal._tcp",  # 35
    "_turn._tcp",  # 33
    "_stun._tcp",  # 33
    "_h323ls._tcp",  # 33
    "_x-puppet._tcp",  # 30
    "_h323cs._udp",  # 27
    "_stuns._tcp",  # 26
    "_jabber-client._tcp",  # 25
    "_x-puppet-ca._tcp",  # 22
    "_ts3._udp",  # 22
    "_minecraft._tcp",  # 22
    "_turns._tcp",  # 21
    "_ldaps._tcp",  # 21
    "_xmpps-client._tcp",  # 20
    "_https._tcp",  # 19
    "_ftp._tcp",  # 19
    "_xmpp-server._udp",  # 18
    "_xmpp-client._udp",  # 17
    "_jabber._udp",  # 17
    "_jabber-client._udp",  # 17
    "_xmpps-server._tcp",  # 15
    "_finger._tcp",  # 14
    "_stuns._udp",  # 12
    "_hkp._tcp",  # 12
    "_vlmcs._udp",  # 11
    "_turns._udp",  # 11
    "_tftp._udp",  # 11
    "_ssh._tcp",  # 11
    "_rtps._udp",  # 11
    "_mysqlsrv._tcp",  # 11
    "_hkps._tcp",  # 11
    "_h323be._udp",  # 11
    "_dns._tcp",  # 11
    "_wss._tcp",  # 10
    "_wpad._tcp",  # 10
    "_whois._tcp",  # 10
    "_webexconnect._tcp",  # 10
    "_webexconnects._tcp",  # 10
    "_vnc._tcp",  # 10
    "_test._tcp",  # 10
    "_telnet._tcp",  # 10
    "_telnets._tcp",  # 10
    "_teamspeak._tcp",  # 10
    "_svns._tcp",  # 10
    "_svcp._tcp",  # 10
    "_smb._tcp",  # 10
    "_sip-tls._tcp",  # 10
    "_sftp._tcp",  # 10
    "_secure-pop3._tcp",  # 10
    "_secure-imap._tcp",  # 10
    "_rtsp._tcp",  # 10
    "_rtps._tcp",  # 10
    "_rpc._tcp",  # 10
    "_rfb._tcp",  # 10
    "_raop._tcp",  # 10
    "_pstn._tcp",  # 10
    "_presence._tcp",  # 10
    "_pkixrep._tcp",  # 10
    "_pgprevokations._tcp",  # 10
    "_pgpkeys._tcp",  # 10
    "_ocsp._tcp",  # 10
    "_nntp._tcp",  # 10
    "_nfs._tcp",  # 10
    "_netbios-ssn._tcp",  # 10
    "_netbios-ns._tcp",  # 10
    "_netbios-dgm._tcp",  # 10
    "_mumble._tcp",  # 10
    "_msrpc._tcp",  # 10
    "_mqtts._tcp",  # 10
    "_minecraft._udp",  # 10
    "_iscsi._tcp",  # 10
    "_ircs._tcp",  # 10
    "_ipp._tcp",  # 10
    "_ipps._tcp",  # 10
    "_h323be._tcp",  # 10
    "_gits._tcp",  # 10
    "_ftps._tcp",  # 10
    "_ftpes._tcp",  # 10
    "_dnss._udp",  # 10
    "_dnss._tcp",  # 10
    "_diameter._tcp",  # 10
    "_crl._tcp",  # 10
    "_crls._tcp",  # 10
    "_cmp._tcp",  # 10
    "_certificates._tcp",  # 10
    "_aix._tcp",  # 10
    "_afpovertcp._tcp",  # 10
    "_collab-edge._tls",  # 6
    "_tcp",  # 5
    "_wildcard",  # 3
    "_client._smtp",  # 3
    "_udp",  # 2
    "_tls",  # 2
    "_msdcs",  # 2
    "_gc._msdcs",  # 2
    "_ldaps._tcp.dc._msdcs",  # 1
    "_kerberos._tcp.kdc._msdcs",  # 1
    "_kerberos.tcp.dc._msdcs",  # 1
    "_imap",  # 1
    "_iax",  # 1
]


class dnscommonsrv(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Check for common SRV records"}
    options = {"top": 50, "max_event_handlers": 10}
    options_desc = {
        "top": "How many of the top SRV records to check",
        "max_event_handlers": "How many instances of the module to run concurrently",
    }
    _max_event_handlers = 10

    def _incoming_dedup_hash(self, event):
        # dedupe by parent
        parent_domain = self.helpers.parent_domain(event.data)
        return hash(parent_domain), "already processed parent domain"

    async def filter_event(self, event):
        # skip SRV wildcards
        if "SRV" in await self.helpers.is_wildcard(event.host):
            return False
        return True

    async def handle_event(self, event):
        top = int(self.config.get("top", 50))
        parent_domain = self.helpers.parent_domain(event.data)
        queries = [f"{srv}.{parent_domain}" for srv in common_srvs[:top]]
        async for query, results in self.helpers.resolve_batch(queries, type="srv"):
            if results:
                await self.emit_event(query, "DNS_NAME", tags=["srv-record"], source=event)



from bbot.modules.templates.shodan import shodan


class shodan_dns(shodan):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query Shodan for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "Shodan API key"}

    base_url = "https://api.shodan.io"

    async def request_url(self, query):
        url = f"{self.base_url}/dns/domain/{self.helpers.quote(query)}?key={self.api_key}"
        response = await self.request_with_fail_count(url)
        return response

    def parse_results(self, r, query):
        json = r.json()
        if json:
            for hostname in json.get("subdomains", []):
                yield f"{hostname}.{query}"



import re
from bbot.modules.base import BaseModule


class ajaxpro(BaseModule):
    """
    Reference: https://mogwailabs.de/en/blog/2022/01/vulnerability-spotlight-rce-in-ajax.net-professional/
    """

    ajaxpro_regex = re.compile(r'<script.+src="([\/a-zA-Z0-9\._]+,[a-zA-Z0-9\._]+\.ashx)"')
    watched_events = ["HTTP_RESPONSE", "URL"]
    produced_events = ["VULNERABILITY", "FINDING"]
    flags = ["active", "safe", "web-thorough"]
    meta = {"description": "Check for potentially vulnerable Ajaxpro instances"}

    async def handle_event(self, event):
        if event.type == "URL":
            if "dir" not in event.tags:
                return False
            for stem in ["ajax", "ajaxpro"]:
                probe_url = f"{event.data}{stem}/whatever.ashx"
                probe = await self.helpers.request(probe_url)
                if probe:
                    if probe.status_code == 200:
                        probe_confirm = await self.helpers.request(f"{event.data}a/whatever.ashx")
                        if probe_confirm:
                            if probe_confirm.status_code != 200:
                                await self.emit_event(
                                    {
                                        "host": str(event.host),
                                        "url": event.data,
                                        "description": f"Ajaxpro Detected (Version Unconfirmed) Trigger: [{probe_url}]",
                                    },
                                    "FINDING",
                                    event,
                                )

        elif event.type == "HTTP_RESPONSE":
            resp_body = event.data.get("body", None)
            if resp_body:
                ajaxpro_regex_result = self.ajaxpro_regex.search(resp_body)
                if ajaxpro_regex_result:
                    ajax_pro_path = ajaxpro_regex_result.group(0)
                    await self.emit_event(
                        {
                            "host": str(event.host),
                            "url": event.data["url"],
                            "description": f"Ajaxpro Detected (Version Unconfirmed) Trigger: [{ajax_pro_path}]",
                        },
                        "FINDING",
                        event,
                    )



from .base import ModuleTestBase

dehashed_domain_response = {
    "balance": 10000,
    "entries": [
        {
            "id": "4363462346",
            "email": "bob@blacklanternsecurity.com",
            "ip_address": "",
            "username": "bob@bob.com",
            "password": "",
            "hashed_password": "$2a$12$pVmwJ7pXEr3mE.DmCCE4fOUDdeadbeefd2KuCy/tq1ZUFyEOH2bve",
            "name": "Bob Smith",
            "vin": "",
            "address": "",
            "phone": "+91283423839",
            "database_name": "eatstreet",
        },
        {
            "id": "234623453454",
            "email": "tim@blacklanternsecurity.com",
            "ip_address": "",
            "username": "timmy",
            "password": "TimTamSlam69",
            "hashed_password": "",
            "name": "Tim Tam",
            "vin": "",
            "address": "",
            "phone": "+123455667",
            "database_name": "eatstreet",
        },
    ],
    "success": True,
    "took": "61µs",
    "total": 2,
}


class TestDehashed(ModuleTestBase):
    modules_overrides = ["dehashed", "speculate"]
    config_overrides = {
        "scope_report_distance": 2,
        "modules": {"dehashed": {"username": "admin", "api_key": "deadbeef"}},
    }

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url=f"https://api.dehashed.com/search?query=domain:blacklanternsecurity.com&size=10000&page=1",
            json=dehashed_domain_response,
        )

    def check(self, module_test, events):
        assert len(events) == 11
        assert 1 == len([e for e in events if e.type == "DNS_NAME" and e.data == "blacklanternsecurity.com"])
        assert 1 == len([e for e in events if e.type == "ORG_STUB" and e.data == "blacklanternsecurity"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "EMAIL_ADDRESS"
                and e.data == "bob@bob.com"
                and e.scope_distance == 1
                and "affiliate" in e.tags
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "bob.com" and e.scope_distance == 1 and "affiliate" in e.tags
            ]
        )
        assert 1 == len([e for e in events if e.type == "EMAIL_ADDRESS" and e.data == "bob@blacklanternsecurity.com"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "USERNAME"
                and e.data == "bob@blacklanternsecurity.com:bob@bob.com"
                and e.source.data == "bob@blacklanternsecurity.com"
            ]
        )
        assert 1 == len([e for e in events if e.type == "EMAIL_ADDRESS" and e.data == "tim@blacklanternsecurity.com"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HASHED_PASSWORD"
                and e.data
                == "bob@blacklanternsecurity.com:$2a$12$pVmwJ7pXEr3mE.DmCCE4fOUDdeadbeefd2KuCy/tq1ZUFyEOH2bve"
            ]
        )
        assert 1 == len(
            [e for e in events if e.type == "PASSWORD" and e.data == "tim@blacklanternsecurity.com:TimTamSlam69"]
        )
        assert 1 == len([e for e in events if e.type == "USERNAME" and e.data == "tim@blacklanternsecurity.com:timmy"])



import re
import random
import string

from bbot.modules.deadly.ffuf import ffuf
from bbot.core.helpers.misc import parse_list_string


def find_common_prefixes(strings, minimum_set_length=4):
    prefix_candidates = [s[:i] for s in strings if len(s) == 6 for i in range(3, 6)]
    frequency_dict = {item: prefix_candidates.count(item) for item in prefix_candidates}
    frequency_dict = {k: v for k, v in frequency_dict.items() if v >= minimum_set_length}
    prefix_list = list(set(frequency_dict.keys()))

    found_prefixes = set()
    for prefix in prefix_list:
        prefix_frequency = frequency_dict[prefix]
        is_substring = False

        for k, v in frequency_dict.items():
            if prefix != k:
                if prefix in k:
                    is_substring = True
        if not is_substring:
            found_prefixes.add(prefix)
        else:
            if prefix_frequency > v and (len(k) - len(prefix) == 1):
                found_prefixes.add(prefix)
    return list(found_prefixes)


class ffuf_shortnames(ffuf):
    watched_events = ["URL_HINT"]
    produced_events = ["URL_UNVERIFIED"]
    flags = ["aggressive", "active", "iis-shortnames", "web-thorough"]
    meta = {"description": "Use ffuf in combination IIS shortnames"}

    options = {
        "wordlist": "",  # default is defined within setup function
        "wordlist_extensions": "",  # default is defined within setup function
        "lines": 1000000,
        "max_depth": 1,
        "version": "2.0.0",
        "extensions": "",
        "ignore_redirects": True,
        "find_common_prefixes": False,
        "find_delimiters": True,
    }

    options_desc = {
        "wordlist": "Specify wordlist to use when finding directories",
        "wordlist_extensions": "Specify wordlist to use when making extension lists",
        "lines": "take only the first N lines from the wordlist when finding directories",
        "max_depth": "the maximum directory depth to attempt to solve",
        "version": "ffuf version",
        "extensions": "Optionally include a list of extensions to extend the keyword with (comma separated)",
        "ignore_redirects": "Explicitly ignore redirects (301,302)",
        "find_common_prefixes": "Attempt to automatically detect common prefixes and make additional ffuf runs against them",
        "find_delimiters": "Attempt to detect common delimiters and make additional ffuf runs against them",
    }

    deps_ansible = [
        {
            "name": "Download ffuf",
            "unarchive": {
                "src": "https://github.com/ffuf/ffuf/releases/download/v#{BBOT_MODULES_FFUF_VERSION}/ffuf_#{BBOT_MODULES_FFUF_VERSION}_#{BBOT_OS_PLATFORM}_#{BBOT_CPU_ARCH}.tar.gz",
                "include": "ffuf",
                "dest": "#{BBOT_TOOLS}",
                "remote_src": True,
            },
        }
    ]

    in_scope_only = True

    async def setup(self):
        self.canary = "".join(random.choice(string.ascii_lowercase) for i in range(10))
        wordlist = self.config.get("wordlist", "")
        if not wordlist:
            wordlist = f"{self.helpers.wordlist_dir}/ffuf_shortname_candidates.txt"
        self.debug(f"Using [{wordlist}] for shortname candidate list")
        self.wordlist = await self.helpers.wordlist(wordlist)
        self.wordlist_lines = list(self.helpers.read_file(self.wordlist))

        wordlist_extensions = self.config.get("wordlist_extensions", "")
        if not wordlist_extensions:
            wordlist_extensions = f"{self.helpers.wordlist_dir}/raft-small-extensions-lowercase_CLEANED.txt"
        self.debug(f"Using [{wordlist_extensions}] for shortname candidate extension list")
        self.wordlist_extensions = await self.helpers.wordlist(wordlist_extensions)

        try:
            self.extensions = parse_list_string(self.config.get("extensions", ""))
            self.debug(f"Using custom extensions: [{','.join(self.extensions)}]")
        except ValueError as e:
            self.warning(f"Error parsing extensions: {e}")
            return False

        self.ignore_redirects = self.config.get("ignore_redirects")

        self.per_host_collection = {}
        self.shortname_to_event = {}
        return True

    def build_extension_list(self, event):
        used_extensions = []
        extension_hint = event.parsed.path.rsplit(".", 1)[1].lower().strip()
        if len(extension_hint) == 3:
            with open(self.wordlist_extensions) as f:
                for l in f:
                    l = l.lower().lstrip(".")
                    if l.lower().startswith(extension_hint):
                        used_extensions.append(l.strip())
            return used_extensions
        else:
            return [extension_hint]

    def find_delimiter(self, hint):
        delimiters = ["_", "-"]
        for d in delimiters:
            if d in hint:
                if not hint.startswith(d) and not hint.endswith(d):
                    return d, hint.split(d)[0], hint.split(d)[1]
        return None

    async def filter_event(self, event):
        return True

    async def handle_event(self, event):
        if event.source.type == "URL":
            filename_hint = re.sub(r"~\d", "", event.parsed.path.rsplit(".", 1)[0].split("/")[-1]).lower()

            host = f"{event.source.parsed.scheme}://{event.source.parsed.netloc}/"
            if host not in self.per_host_collection.keys():
                self.per_host_collection[host] = [(filename_hint, event.source.data)]

            else:
                self.per_host_collection[host].append((filename_hint, event.source.data))

            self.shortname_to_event[filename_hint] = event

            root_stub = "/".join(event.parsed.path.split("/")[:-1])
            root_url = f"{event.parsed.scheme}://{event.parsed.netloc}{root_stub}/"

            if "shortname-file" in event.tags:
                used_extensions = self.build_extension_list(event)

            if len(filename_hint) == 6:
                tempfile, tempfile_len = self.generate_templist(prefix=filename_hint)
                self.verbose(
                    f"generated temp word list of size [{str(tempfile_len)}] for filename hint: [{filename_hint}]"
                )

            else:
                tempfile = self.helpers.tempfile([filename_hint], pipe=False)
                tempfile_len = 1

            if tempfile_len > 0:
                if "shortname-file" in event.tags:
                    for ext in used_extensions:
                        async for r in self.execute_ffuf(tempfile, root_url, suffix=f".{ext}"):
                            await self.emit_event(
                                r["url"], "URL_UNVERIFIED", source=event, tags=[f"status-{r['status']}"]
                            )

                elif "shortname-directory" in event.tags:
                    async for r in self.execute_ffuf(tempfile, root_url, exts=["/"]):
                        r_url = f"{r['url'].rstrip('/')}/"
                        await self.emit_event(r_url, "URL_UNVERIFIED", source=event, tags=[f"status-{r['status']}"])

            if self.config.get("find_delimiters"):
                if "shortname-directory" in event.tags:
                    delimiter_r = self.find_delimiter(filename_hint)
                    if delimiter_r:
                        delimiter, prefix, partial_hint = delimiter_r
                        self.verbose(f"Detected delimiter [{delimiter}] in hint [{filename_hint}]")
                        tempfile, tempfile_len = self.generate_templist(prefix=partial_hint)
                        async for r in self.execute_ffuf(
                            tempfile, root_url, prefix=f"{prefix}{delimiter}", exts=["/"]
                        ):
                            await self.emit_event(
                                r["url"], "URL_UNVERIFIED", source=event, tags=[f"status-{r['status']}"]
                            )

                elif "shortname-file" in event.tags:
                    for ext in used_extensions:
                        delimiter_r = self.find_delimiter(filename_hint)
                        if delimiter_r:
                            delimiter, prefix, partial_hint = delimiter_r
                            self.verbose(f"Detected delimiter [{delimiter}] in hint [{filename_hint}]")
                            tempfile, tempfile_len = self.generate_templist(prefix=partial_hint)
                            async for r in self.execute_ffuf(
                                tempfile, root_url, prefix=f"{prefix}{delimiter}", suffix=f".{ext}"
                            ):
                                await self.emit_event(
                                    r["url"], "URL_UNVERIFIED", source=event, tags=[f"status-{r['status']}"]
                                )

    async def finish(self):
        if self.config.get("find_common_prefixes"):
            per_host_collection = dict(self.per_host_collection)
            self.per_host_collection.clear()

            for host, hint_tuple_list in per_host_collection.items():
                hint_list = [x[0] for x in hint_tuple_list]

                common_prefixes = find_common_prefixes(hint_list)
                for prefix in common_prefixes:
                    self.verbose(f"Found common prefix: [{prefix}] for host [{host}]")
                    for hint_tuple in hint_tuple_list:
                        hint, url = hint_tuple
                        if hint.startswith(prefix):
                            partial_hint = hint[len(prefix) :]

                            # safeguard to prevent loading the entire wordlist
                            if len(partial_hint) > 0:
                                tempfile, tempfile_len = self.generate_templist(prefix=partial_hint)

                                if "shortname-directory" in self.shortname_to_event[hint].tags:
                                    self.verbose(
                                        f"Running common prefix check for URL_HINT: {hint} with prefix: {prefix} and partial_hint: {partial_hint}"
                                    )

                                    async for r in self.execute_ffuf(tempfile, url, prefix=prefix, exts=["/"]):
                                        await self.emit_event(
                                            r["url"],
                                            "URL_UNVERIFIED",
                                            source=self.shortname_to_event[hint],
                                            tags=[f"status-{r['status']}"],
                                        )
                                elif "shortname-file" in self.shortname_to_event[hint].tags:
                                    used_extensions = self.build_extension_list(self.shortname_to_event[hint])

                                    for ext in used_extensions:
                                        self.verbose(
                                            f"Running common prefix check for URL_HINT: {hint} with prefix: {prefix}, extension: .{ext}, and partial_hint: {partial_hint}"
                                        )
                                        async for r in self.execute_ffuf(
                                            tempfile, url, prefix=prefix, suffix=f".{ext}"
                                        ):
                                            await self.emit_event(
                                                r["url"],
                                                "URL_UNVERIFIED",
                                                source=self.shortname_to_event[hint],
                                                tags=[f"status-{r['status']}"],
                                            )



from .base import ModuleTestBase


class TestDNSCommonSRV(ModuleTestBase):
    targets = ["blacklanternsecurity.notreal"]
    config_overrides = {"dns_resolution": True}

    async def setup_after_prep(self, module_test):
        module_test.mock_dns(
            {
                "_ldap._tcp.gc._msdcs.blacklanternsecurity.notreal": {
                    "SRV": ["0 100 3268 asdf.blacklanternsecurity.notreal"]
                },
                "asdf.blacklanternsecurity.notreal": {"A": "1.2.3.4"},
            }
        )

    def check(self, module_test, events):
        assert any(
            e.data == "_ldap._tcp.gc._msdcs.blacklanternsecurity.notreal" for e in events
        ), "Failed to detect subdomain"
        assert any(e.data == "asdf.blacklanternsecurity.notreal" for e in events), "Failed to detect subdomain"
        assert not any(e.data == "_ldap._tcp.dc._msdcs.blacklanternsecurity.notreal" for e in events), "False positive"
